from abc import ABC, abstractmethod
import re
from typing import (
    ClassVar,
    Optional,
    Union,
    List,
    Sequence,
    Dict,
    Type,
    get_origin,
    get_args,
    get_type_hints,
)
from collections.abc import Sequence as SequenceABC
from base64 import b64encode
from sigma.types import (
    Placeholder,
    SigmaBool,
    SigmaCasedString,
    SigmaExists,
    SigmaExpansion,
    SigmaFieldReference,
    SigmaRegularExpressionFlag,
    SigmaType,
    SigmaString,
    SigmaNumber,
    SpecialChars,
    SigmaRegularExpression,
    SigmaCompareExpression,
    SigmaCIDRExpression,
)
from sigma.conditions import ConditionAND
from sigma.exceptions import SigmaRuleLocation, SigmaTypeError, SigmaValueError
import sigma


### Base Classes ###
class SigmaModifier(ABC):
    """Base class for all Sigma modifiers"""

    detection_item: "sigma.rule.SigmaDetectionItem"
    applied_modifiers: List["SigmaModifier"]

    def __init__(
        self,
        detection_item: "sigma.rule.SigmaDetectionItem",
        applied_modifiers: List["SigmaModifier"],
        source: Optional[SigmaRuleLocation] = None,
    ):
        self.detection_item = detection_item
        self.applied_modifiers = applied_modifiers
        self.source = source

    def type_check(self, val: Union[SigmaType, Sequence[SigmaType]], explicit_type=None) -> bool:
        th = (
            explicit_type or get_type_hints(self.modify)["val"]
        )  # get type annotation from val parameter of apply method or explicit_type parameter
        to = get_origin(th)  # get possible generic type of type hint
        if to is None:  # Plain type in annotation
            return isinstance(val, th)
        elif to is Union:  # type hint is Union of multiple types, check if val is one of them
            for t in get_args(th):
                if isinstance(val, t):
                    return True
            return False
        elif to is SequenceABC:  # type hint is sequence
            inner_type = get_args(th)[0]
            return all([self.type_check(item, explicit_type=inner_type) for item in val])

    @abstractmethod
    def modify(
        self, val: Union[SigmaType, Sequence[SigmaType]]
    ) -> Union[SigmaType, List[SigmaType]]:
        """This method should be overridden with the modifier implementation."""

    def apply(self, val: Union[SigmaType, Sequence[SigmaType]]) -> List[SigmaType]:
        """
        Modifier entry point containing the default operations:
        * Type checking
        * Ensure returned value is a list
        * Handle values of SigmaExpansion objects separately.
        """
        if isinstance(val, SigmaExpansion):  # Handle each SigmaExpansion item separately
            return [SigmaExpansion([va for v in val.values for va in self.apply(v)])]
        else:
            if not self.type_check(val):
                raise SigmaTypeError(
                    f"Modifier {self.__class__.__name__} incompatible to value type of '{ val }'",
                    source=self.source,
                )
            r = self.modify(val)
            if isinstance(r, List):
                return r
            else:
                return [r]


class SigmaValueModifier(SigmaModifier):
    """Base class for all modifiers that handle each value for the modifier scope separately"""

    @abstractmethod
    def modify(self, val: SigmaType) -> Union[SigmaType, List[SigmaType]]:
        """This method should be overridden with the modifier implementation."""


class SigmaListModifier(SigmaModifier):
    """Base class for all modifiers that handle all values for the modifier scope as a whole."""

    @abstractmethod
    def modify(self, val: Sequence[SigmaType]) -> Union[SigmaType, List[SigmaType]]:
        """This method should be overridden with the modifier implementation."""


### Modifier Implementations ###
class SigmaContainsModifier(SigmaValueModifier):
    """Puts wildcards around a string to match it somewhere inside another string instead of as a whole."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression]
    ) -> Union[SigmaString, SigmaRegularExpression]:
        if isinstance(val, SigmaString):
            if not val.startswith(SpecialChars.WILDCARD_MULTI):
                val = SpecialChars.WILDCARD_MULTI + val
            if not val.endswith(SpecialChars.WILDCARD_MULTI):
                val += SpecialChars.WILDCARD_MULTI
        elif isinstance(val, SigmaRegularExpression):
            if val.regexp[:2] != ".*" and val.regexp[0] != "^":
                val.regexp = ".*" + val.regexp
            if val.regexp[-2:] != ".*" and val.regexp[-1] != "$":
                val.regexp = val.regexp + ".*"
            val.compile()
        return val


class SigmaStartswithModifier(SigmaValueModifier):
    """Puts a wildcard at the end of a string to match arbitrary values after the given prefix."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression]
    ) -> Union[SigmaString, SigmaRegularExpression]:
        if isinstance(val, SigmaString):
            if not val.endswith(SpecialChars.WILDCARD_MULTI):
                val += SpecialChars.WILDCARD_MULTI
        elif isinstance(val, SigmaRegularExpression):
            if val.regexp[-2:] != ".*" and val.regexp[-1] != "$":
                val.regexp = val.regexp + ".*"
            val.compile()
        return val


class SigmaEndswithModifier(SigmaValueModifier):
    """Puts a wildcard before a string to match arbitrary values before it."""

    def modify(
        self, val: Union[SigmaString, SigmaRegularExpression]
    ) -> Union[SigmaString, SigmaRegularExpression]:
        if isinstance(val, SigmaString):
            if not val.startswith(SpecialChars.WILDCARD_MULTI):
                val = SpecialChars.WILDCARD_MULTI + val
        elif isinstance(val, SigmaRegularExpression):
            if val.regexp[:2] != ".*" and val.regexp[0] != "^":
                val.regexp = ".*" + val.regexp
            val.compile()
        return val


class SigmaBase64Modifier(SigmaValueModifier):
    """Encode string as Base64 value."""

    def modify(self, val: SigmaString) -> SigmaString:
        if val.contains_special():
            raise SigmaValueError(
                "Base64 encoding of strings with wildcards is not allowed",
                source=self.source,
            )
        return SigmaString(b64encode(bytes(val)).decode())


class SigmaBase64OffsetModifier(SigmaValueModifier):
    """
    Encode string as Base64 value with different offsets to match it at different locations in
    encoded form.
    """

    start_offsets = (0, 2, 3)
    end_offsets = (None, -3, -2)

    def modify(self, val: SigmaString) -> SigmaExpansion:
        if val.contains_special():
            raise SigmaValueError(
                "Base64 encoding of strings with wildcards is not allowed",
                source=self.source,
            )
        return SigmaExpansion(
            [
                SigmaString(
                    b64encode(i * b" " + bytes(val))[
                        self.start_offsets[i] : self.end_offsets[(len(val) + i) % 3]
                    ].decode()
                )
                for i in range(3)
            ]
        )


class SigmaWideModifier(SigmaValueModifier):
    """Encode string as wide string (UTF-16LE)."""

    def modify(self, val: SigmaString) -> SigmaString:
        r = list()
        for item in val.s:
            if isinstance(
                item, str
            ):  # put 0x00 after each character by encoding it to utf-16le and decoding it as utf-8
                try:
                    r.append(item.encode("utf-16le").decode("utf-8"))
                except UnicodeDecodeError:  # this method only works for ascii characters
                    raise SigmaValueError(
                        f"Wide modifier only allowed for ascii strings, input string '{str(val)}' isn't one",
                        source=self.source,
                    )
            else:  # just append special characters without further handling
                r.append(item)

        s = SigmaString()
        s.s = tuple(r)
        return s


class SigmaWindowsDashModifier(SigmaValueModifier):
    """
    Expand parameter characters / and - that are often interchangeable in Windows into the other
    form if it appears between word boundaries. E.g. in -param-name the first dash will be expanded
    into /param-name while the second dash is left untouched.
    """

    def modify(self, val: SigmaString) -> SigmaExpansion:
        def callback(p: Placeholder):
            if p.name == "_windash":
                yield from ("-", "/")
            else:
                yield p

        return SigmaExpansion(
            val.replace_with_placeholder(re.compile("\\B[-/]\\b"), "_windash").replace_placeholders(
                callback
            )
        )


class SigmaRegularExpressionModifier(SigmaValueModifier):
    """Treats string value as (case-sensitive) regular expression."""

    def modify(self, val: SigmaString) -> SigmaRegularExpression:
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "Regular expression modifier only applicable to unmodified values",
                source=self.source,
            )
        return SigmaRegularExpression(val.original)


class SigmaRegularExpressionFlagModifier(SigmaValueModifier):
    """Generic base class for setting a regular expression flag including checks"""

    flag: ClassVar[SigmaRegularExpressionFlag]

    def modify(self, val: SigmaRegularExpression) -> SigmaRegularExpression:
        val.add_flag(self.flag)
        return val


class SigmaRegularExpressionIgnoreCaseFlagModifier(SigmaRegularExpressionFlagModifier):
    """Match regular expression case-insensitive."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.IGNORECASE


class SigmaRegularExpressionMultilineFlagModifier(SigmaRegularExpressionFlagModifier):
    """Match regular expression across multiple lines."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.MULTILINE


class SigmaRegularExpressionDotAllFlagModifier(SigmaRegularExpressionFlagModifier):
    """Regular expression dot matches all characters."""

    flag: ClassVar[SigmaRegularExpressionFlag] = SigmaRegularExpressionFlag.DOTALL


class SigmaCaseSensitiveModifier(SigmaValueModifier):
    def modify(self, val: SigmaString) -> SigmaCasedString:
        return SigmaCasedString.from_sigma_string(val)


class SigmaCIDRModifier(SigmaValueModifier):
    """Treat value as IP (v4 or v6) CIDR network."""

    def modify(self, val: SigmaString) -> SigmaCIDRExpression:
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "CIDR expression modifier only applicable to unmodified values",
                source=self.source,
            )
        return SigmaCIDRExpression(str(val), source=self.source)


class SigmaAllModifier(SigmaListModifier):
    """Match all values of a list instead of any pf them."""

    def modify(self, val: Sequence[SigmaType]) -> List[SigmaType]:
        self.detection_item.value_linking = ConditionAND
        return val


class SigmaCompareModifier(SigmaValueModifier):
    """Base class for numeric comparison operator modifiers."""

    op: ClassVar[SigmaCompareExpression.CompareOperators]

    def modify(self, val: SigmaNumber) -> SigmaCompareExpression:
        return SigmaCompareExpression(val, self.op, self.source)


class SigmaLessThanModifier(SigmaCompareModifier):
    """Numeric less than (<) matching."""

    op: ClassVar[
        SigmaCompareExpression.CompareOperators
    ] = SigmaCompareExpression.CompareOperators.LT


class SigmaLessThanEqualModifier(SigmaCompareModifier):
    """Numeric less than or equal (<=) matching."""

    op: ClassVar[
        SigmaCompareExpression.CompareOperators
    ] = SigmaCompareExpression.CompareOperators.LTE


class SigmaGreaterThanModifier(SigmaCompareModifier):
    """Numeric greater than (>) matching."""

    op: ClassVar[
        SigmaCompareExpression.CompareOperators
    ] = SigmaCompareExpression.CompareOperators.GT


class SigmaGreaterThanEqualModifier(SigmaCompareModifier):
    """Numeric greater than or equal (>=) matching."""

    op: ClassVar[
        SigmaCompareExpression.CompareOperators
    ] = SigmaCompareExpression.CompareOperators.GTE


class SigmaFieldReferenceModifier(SigmaValueModifier):
    """Modifiers a plain string into the field reference type."""

    def modify(self, val: SigmaString) -> SigmaFieldReference:
        if val.contains_special():
            raise SigmaValueError("Field references must not contain wildcards", source=self.source)
        return SigmaFieldReference(val.to_plain())


class SigmaExistsModifier(SigmaValueModifier):
    """Modifies to check if the field name provided as value exists in the matched event."""

    def modify(self, val: SigmaBool) -> SigmaExists:
        if self.detection_item.field is None:
            raise SigmaValueError("Exists modifier must be applied to field", source=self.source)
        if len(self.applied_modifiers) > 0:
            raise SigmaValueError(
                "Exists modifier only applicable to unmodified boolean values",
                source=self.source,
            )
        return SigmaExists(val.boolean)


class SigmaExpandModifier(SigmaValueModifier):
    """
    Modifier for expansion of placeholders in values. It replaces placeholder strings (%something%)
    with stub objects that are later expanded to one or multiple strings or replaced with some SIEM
    specific list item or lookup by the processing pipeline.
    """

    def modify(self, val: SigmaString) -> SigmaString:
        return val.insert_placeholders()


# Mapping from modifier identifier strings to modifier classes
modifier_mapping: Dict[str, Type[SigmaModifier]] = {
    "contains": SigmaContainsModifier,
    "startswith": SigmaStartswithModifier,
    "endswith": SigmaEndswithModifier,
    "exists": SigmaExistsModifier,
    "base64": SigmaBase64Modifier,
    "base64offset": SigmaBase64OffsetModifier,
    "wide": SigmaWideModifier,
    "windash": SigmaWindowsDashModifier,
    "re": SigmaRegularExpressionModifier,
    "i": SigmaRegularExpressionIgnoreCaseFlagModifier,
    "ignorecase": SigmaRegularExpressionIgnoreCaseFlagModifier,
    "m": SigmaRegularExpressionMultilineFlagModifier,
    "multiline": SigmaRegularExpressionMultilineFlagModifier,
    "s": SigmaRegularExpressionDotAllFlagModifier,
    "dotall": SigmaRegularExpressionDotAllFlagModifier,
    "cased": SigmaCaseSensitiveModifier,
    "cidr": SigmaCIDRModifier,
    "all": SigmaAllModifier,
    "lt": SigmaLessThanModifier,
    "lte": SigmaLessThanEqualModifier,
    "gt": SigmaGreaterThanModifier,
    "gte": SigmaGreaterThanEqualModifier,
    "fieldref": SigmaFieldReferenceModifier,
    "expand": SigmaExpandModifier,
}

# Mapping from modifier class to identifier
reverse_modifier_mapping: Dict[str, str] = {
    modifier_class.__name__: identifier for identifier, modifier_class in modifier_mapping.items()
}



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import sigma
from sigma.types import SigmaFieldReference, SigmaString, SigmaType, sigma_type
from typing import Dict, List, Pattern, Literal, Optional, Union
import re
from sigma.rule import SigmaDetection, SigmaRule, SigmaDetectionItem, SigmaLogSource
from sigma.exceptions import SigmaConfigurationError, SigmaRegularExpressionError


### Base Classes ###
class ProcessingCondition(ABC):
    """Anchor base class for all processing condition types."""


@dataclass
class RuleProcessingCondition(ProcessingCondition, ABC):
    """
    Base for Sigma rule processing condition classes used in processing pipelines.
    """

    @abstractmethod
    def match(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule
    ) -> bool:
        """Match condition on Sigma rule."""


class FieldNameProcessingCondition(ProcessingCondition, ABC):
    """
    Base class for conditions on field names in detection items, Sigma rule field lists and other
    use cases that require matching on field names without detection item context.
    """

    @abstractmethod
    def match_field_name(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", field: str
    ) -> bool:
        "The method match is called for each field name and must return a bool result."

    def match_detection_item(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        """
        Field names can be contained in the detection item field as well as in field references in
        detection item values. The detection item matching returns True for both cases, but in
        subsequent processing it has to be verified which part of the detection item has matched and
        should be subject of processing actions (e.g. field name mapping). This can be done with the
        methods

        * `match_detection_item_field` for the field of a detection item
        * `match_detection_item_value` for the whole value list of a detection item and
        * `match_value` for single detection items values.
        """
        return self.match_detection_item_field(
            pipeline, detection_item
        ) or self.match_detection_item_value(pipeline, detection_item)

    def match_detection_item_field(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        """Returns True if the field of the detection item matches the implemented field name condition."""
        return self.match_field_name(pipeline, detection_item.field)

    def match_detection_item_value(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        """Returns True if any value of a detection item contains a field reference to a field name
        matching the implemented field name condition. Processing actions must only be applied to
        matching individual values determined by `match_value`."""
        return any((self.match_value(pipeline, value) for value in detection_item.value))

    def match_value(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", value: SigmaType
    ) -> bool:
        """
        Checks if a detection item value matches the field name condition implemented in
        `match_field_name` if it is a field reference. For all other types the method returns False.
        """
        if isinstance(value, SigmaFieldReference):
            return self.match_field_name(pipeline, value.field)
        else:
            return False


@dataclass
class DetectionItemProcessingCondition(ProcessingCondition, ABC):
    """
    Base for Sigma detection item processing condition classes used in processing pipelines.
    """

    @abstractmethod
    def match(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        """Match condition on Sigma rule."""


@dataclass
class ValueProcessingCondition(DetectionItemProcessingCondition):
    """
    Base class for conditions on values in detection items. The 'cond' parameter determines if any or all
    values of a multivalued detection item must match to result in an overall match.

    The method match_value is called for each value and must return a bool result. It should reject values
    which are incompatible with the condition with a False return value.
    """

    cond: Literal["any", "all"]

    def __post_init__(self):
        if self.cond == "any":
            self.match_func = any
        elif self.cond == "all":
            self.match_func = all
        else:
            raise SigmaConfigurationError(
                f"The value '{self.cond}' for the 'cond' parameter is invalid. It must be 'any' or 'all'."
            )

    def match(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        return self.match_func(
            (self.match_value(pipeline, value) for value in detection_item.value)
        )

    @abstractmethod
    def match_value(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", value: SigmaType
    ) -> bool:
        """Match condition on detection item values."""


### Rule Condition Classes ###
@dataclass
class LogsourceCondition(RuleProcessingCondition):
    """
    Matches log source on rule. Not specified log source fields are ignored.
    """

    category: Optional[str] = field(default=None)
    product: Optional[str] = field(default=None)
    service: Optional[str] = field(default=None)

    def __post_init__(self):
        self.logsource = SigmaLogSource(self.category, self.product, self.service)

    def match(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule
    ) -> bool:
        return rule.logsource in self.logsource


@dataclass
class RuleContainsDetectionItemCondition(RuleProcessingCondition):
    """Returns True if rule contains a detection item that matches the given field name and value."""

    field: Optional[str]
    value: Union[str, int, float, bool]

    def __post_init__(self):
        self.sigma_value = sigma_type(self.value)

    def match(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule
    ) -> bool:
        for detection in rule.detection.detections.values():
            if self.find_detection_item(detection):
                return True
        return False

    def find_detection_item(self, detection: Union[SigmaDetectionItem, SigmaDetection]) -> bool:
        if isinstance(detection, SigmaDetection):
            for detection_item in detection.detection_items:
                if self.find_detection_item(detection_item):
                    return True
        elif isinstance(detection, SigmaDetectionItem):
            if (
                detection.field is not None
                and detection.field == self.field
                and self.sigma_value
                in [v for v in detection.value if type(self.sigma_value) == type(v)]
            ):
                return True
        else:
            raise TypeError("Parameter of type SigmaDetection or SigmaDetectionItem expected.")

        return False


@dataclass
class RuleProcessingItemAppliedCondition(RuleProcessingCondition):
    """
    Checks if processing item was applied to rule.
    """

    processing_item_id: str

    def match(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule
    ) -> bool:
        return rule.was_processed_by(self.processing_item_id)


### Field Name Condition Classes ###
@dataclass
class IncludeFieldCondition(FieldNameProcessingCondition):
    """
    Matches on field name if it is contained in fields list. The parameter 'type' determines if field names are matched as
    plain string ("plain") or regular expressions ("re").
    """

    fields: List[str]
    type: Literal["plain", "re"] = field(default="plain")
    patterns: List[Pattern] = field(init=False, repr=False, default_factory=list)

    def __post_init__(self):
        """
        Check if type is known and pre-compile regular expressions.
        """
        if self.type == "plain":
            pass
        elif self.type == "re":
            self.patterns = [re.compile(field) for field in self.fields]
        else:
            raise SigmaConfigurationError(
                f"Invalid detection item field name condition type '{self.type}', supported types are 'plain' or 're'."
            )

    def match_field_name(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        field: Optional[str],
    ) -> bool:
        if field is None:
            return False
        elif self.type == "plain":
            return field in self.fields
        else:  # regular expression matching
            try:
                return any((pattern.match(field) for pattern in self.patterns))
            except Exception as e:
                msg = f" (while processing field '{field}'"
                if len(e.args) > 1:
                    e.args = (e.args[0] + msg,) + e.args[1:]
                else:
                    e.args = (e.args[0] + msg,)
                raise


@dataclass
class ExcludeFieldCondition(IncludeFieldCondition):
    """Matches on field name if it is not contained in fields list."""

    def match_field_name(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        return not super().match_field_name(pipeline, detection_item)


### Detection Item Condition Classes ###
@dataclass
class MatchStringCondition(ValueProcessingCondition):
    """
    Match string values with a regular expression 'pattern'. The parameter 'cond' determines for detection items with multiple
    values if any or all strings must match. Generally, values which aren't strings are skipped in any mode or result in a
    false result in all match mode.
    """

    pattern: str
    negate: bool = False

    def __post_init__(self):
        super().__post_init__()
        try:
            self.re = re.compile(self.pattern)
        except re.error as e:
            raise SigmaRegularExpressionError(
                f"Regular expression '{self.pattern}' is invalid: {str(e)}"
            ) from e

    def match_value(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", value: SigmaType
    ) -> bool:
        if isinstance(value, SigmaString):
            result = self.re.match(str(value))
        else:
            result = False

        if self.negate:
            return not result
        else:
            return result


@dataclass
class DetectionItemProcessingItemAppliedCondition(DetectionItemProcessingCondition):
    """
    Checks if processing item was applied to detection item.
    """

    processing_item_id: str

    def match(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ) -> bool:
        return detection_item.was_processed_by(self.processing_item_id)


@dataclass
class FieldNameProcessingItemAppliedCondition(FieldNameProcessingCondition):
    """
    Checks if processing item was applied to a field name.
    """

    processing_item_id: str

    def match_field_name(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", field: str
    ) -> bool:
        return pipeline.field_was_processed_by(field, self.processing_item_id)

    def match_detection_item(
        self,
        pipeline: "sigma.processing.pipeline.ProcessingPipeline",
        detection_item: SigmaDetectionItem,
    ):
        return detection_item.was_processed_by(self.processing_item_id)


### Condition mappings between rule identifier and class

rule_conditions: Dict[str, RuleProcessingCondition] = {
    "logsource": LogsourceCondition,
    "contains_detection_item": RuleContainsDetectionItemCondition,
    "processing_item_applied": RuleProcessingItemAppliedCondition,
}
detection_item_conditions: Dict[str, DetectionItemProcessingCondition] = {
    "match_string": MatchStringCondition,
    "processing_item_applied": DetectionItemProcessingItemAppliedCondition,
}
field_name_conditions: Dict[str, DetectionItemProcessingCondition] = {
    "include_fields": IncludeFieldCondition,
    "exclude_fields": ExcludeFieldCondition,
    "processing_item_applied": FieldNameProcessingItemAppliedCondition,
}



import re
from collections import Counter
from collections import defaultdict
from dataclasses import dataclass
from typing import ClassVar, Dict, List, Set
from uuid import UUID

from sigma.rule import SigmaRule
from sigma.validators.base import (
    SigmaRuleValidator,
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
)


def is_uuid_v4(val: str) -> bool:
    try:
        id = UUID(str(val))
        if id.version == 4:
            return True
        else:
            return False
    except ValueError:
        return False


@dataclass
class IdentifierExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no identifier (UUID)"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class IdentifierExistenceValidator(SigmaRuleValidator):
    """Checks if rule has identifier."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.id is None:
            return [IdentifierExistenceIssue([rule])]
        else:
            return []


@dataclass
class IdentifierCollisionIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule identifier used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    identifier: UUID


class IdentifierUniquenessValidator(SigmaRuleValidator):
    """Check rule UUID uniqueness."""

    ids: Dict[UUID, List[SigmaRule]]

    def __init__(self):
        self.ids = defaultdict(list)

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.id is not None:
            self.ids[rule.id].append(rule)
        return []

    def finalize(self) -> List[SigmaValidationIssue]:
        return [
            IdentifierCollisionIssue(rules, id) for id, rules in self.ids.items() if len(rules) > 1
        ]


@dataclass
class TitleLengthSigmaHQIssue(SigmaValidationIssue):
    description = "Rule has a title longer than 110 characters"
    severity = SigmaValidationIssueSeverity.MEDIUM


class TitleLengthSigmaHQValidator(SigmaRuleValidator):
    """Checks if rule has a title length longer than 110."""

    def validate(self, rule: SigmaRule) -> List[TitleLengthSigmaHQIssue]:
        if len(rule.title) > 110:
            return [TitleLengthSigmaHQIssue([rule])]
        else:
            return []


@dataclass
class DuplicateTitleIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule title used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    title: str


class DuplicateTitleValidator(SigmaRuleValidator):
    """Check rule title uniqueness."""

    titles: Dict[str, List[SigmaRule]]

    def __init__(self):
        self.titles = defaultdict(list)

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.title is not None:
            self.titles[rule.title].append(rule)
        return []

    def finalize(self) -> List[SigmaValidationIssue]:
        return [
            DuplicateTitleIssue(rules, title)
            for title, rules in self.titles.items()
            if len(rules) > 1
        ]


@dataclass
class DuplicateReferencesIssue(SigmaValidationIssue):
    description: ClassVar[str] = "The same references appears multiple times"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    reference: str


class DuplicateReferencesValidator(SigmaRuleValidator):
    """Validate rule References uniqueness."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        references = Counter(rule.references)
        return [
            DuplicateReferencesIssue([rule], reference)
            for reference, count in references.items()
            if count > 1
        ]


@dataclass
class StatusExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no status"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class StatusExistenceValidator(SigmaRuleValidator):
    """Checks if rule has a status."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.status is None:
            return [StatusExistenceIssue([rule])]
        else:
            return []


@dataclass
class StatusUnsupportedIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has UNSUPPORTED status"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class StatusUnsupportedValidator(SigmaRuleValidator):
    """Checks if rule has a status UNSUPPORTED."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.status and rule.status.name == "UNSUPPORTED":
            return [StatusUnsupportedIssue([rule])]
        else:
            return []


@dataclass
class DateExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no date"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class DateExistenceValidator(SigmaRuleValidator):
    """Checks if rule has a data."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.date is None:
            return [DateExistenceIssue([rule])]
        else:
            return []


@dataclass
class DuplicateFilenameIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule filemane used by multiple rules"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    filename: str


class DuplicateFilenameValidator(SigmaRuleValidator):
    """Check rule filename uniqueness."""

    filenames: Dict[str, List[SigmaRule]]

    def __init__(self):
        self.filenames = defaultdict(list)

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.source is not None:
            self.filenames[rule.source.path.name].append(rule)
        return []

    def finalize(self) -> List[SigmaValidationIssue]:
        return [
            DuplicateFilenameIssue(rules, filename)
            for filename, rules in self.filenames.items()
            if len(rules) > 1
        ]


@dataclass
class FilenameSigmahqIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule filemane doesn't match SigmaHQ standard"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    filename: str


class FilenameSigmahqValidator(SigmaRuleValidator):
    """Check rule filename match SigmaHQ standard."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        filename_pattern = re.compile(r"[a-z0-9_]{10,90}\.yml")
        if rule.source is not None:
            filename = rule.source.path.name
            if filename_pattern.match(filename) is None or not "_" in filename:
                return [FilenameSigmahqIssue(rule, filename)]
        return []


@dataclass
class FilenameLenghIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule filename is too short or long"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    filename: str


class FilenameLenghValidator(SigmaRuleValidator):
    """Check rule filename lengh"""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.source is not None:
            filename = rule.source.path.name
            if len(filename) < 10 or len(filename) > 90:
                return [FilenameLenghIssue(rule, filename)]
        return []


@dataclass
class CustomAttributesIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule use optional field name similar to legit"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    fieldname: str


class CustomAttributesValidator(SigmaRuleValidator):
    """Check if field name is similar to legit one"""

    known_custom_attributes: Set[str] = {
        "realted",
        "relatde",
        "relted",
        "rlated",
        "reference",
    }

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.custom_attributes is not None:
            for k in rule.custom_attributes.keys():
                if k in self.known_custom_attributes:
                    return [CustomAttributesIssue(rule, k)]
        return []


@dataclass
class DescriptionExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no description"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class DescriptionExistenceValidator(SigmaRuleValidator):
    """Checks if rule has a description."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.description is None:
            return [DescriptionExistenceIssue([rule])]
        else:
            return []


@dataclass
class DescriptionLengthIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has a too short description"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class DescriptionLengthValidator(SigmaRuleValidator):
    """Checks if rule has a description."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.description is not None and len(rule.description) < 16:
            return [DescriptionLengthIssue([rule])]
        else:
            return []


@dataclass
class LevelExistenceIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule has no level"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class LevelExistenceValidator(SigmaRuleValidator):
    """Checks if rule has a level."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if rule.level is None:
            return [LevelExistenceIssue([rule])]
        else:
            return []



from dataclasses import dataclass, field
from sigma.conversion.deferred import DeferredQueryExpression
from typing import Any, List, Mapping


@dataclass
class ConversionState:
    """
    State class which is passed as object to each conversion method in query conversion and
    finalization phase. All state information that is required in a later phase of the converison
    should be stored in this class.

    The base class implements deferred query expressions, which are generated in the finalization
    phase. Conversion state objects are initialized at beginning of the conversion of a rule in
    the backends convert_rule method and are discarded after this method finishes, short after
    execution of the finalize_query method.

    In addition to deferred query expressions, ConversionState objects can be used to pass state
    down or up the parse tree.
    """

    deferred: List[DeferredQueryExpression] = field(default_factory=list)
    processing_state: Mapping[str, Any] = field(default_factory=dict)

    def add_deferred_expression(self, deferred: DeferredQueryExpression) -> None:
        self.deferred.append(deferred)

    def has_deferred(self) -> bool:
        """Return True when deferred expressions are contained in state object."""
        return len(self.deferred) > 0



from dataclasses import dataclass
from typing import ClassVar, List
from sigma.modifiers import (
    SigmaContainsModifier,
    SigmaEndswithModifier,
    SigmaStartswithModifier,
)
from sigma.rule import SigmaDetectionItem
from sigma.types import SigmaString, SpecialChars
from sigma.validators.base import (
    SigmaDetectionItemValidator,
    SigmaStringValueValidator,
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
)


@dataclass
class DoubleWildcardIssue(SigmaValidationIssue):
    description: ClassVar[str] = "String contains multiple consecutive * wildcards"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    string: SigmaString


class DoubleWildcardValidator(SigmaStringValueValidator):
    """Check strings for consecutive multi-character wildcards."""

    def validate_value(self, value: SigmaString) -> List[SigmaValidationIssue]:
        prev_wildcard = False
        for c in value.s:
            if c == SpecialChars.WILDCARD_MULTI:
                if prev_wildcard:  # previous character was also a wildcard
                    return [DoubleWildcardIssue([self.rule], value)]
                else:
                    prev_wildcard = True
            else:
                prev_wildcard = False
        return []


@dataclass
class NumberAsStringIssue(SigmaValidationIssue):
    description: ClassVar[str] = "A number was expressed as string"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    string: SigmaString


class NumberAsStringValidator(SigmaStringValueValidator):
    """Check numbers that were expressed as strings."""

    def validate_value(self, value: SigmaString) -> List[SigmaValidationIssue]:
        if len(value.s) == 1 and isinstance(value.s[0], str) and not " " in value.s[0]:
            try:
                int(value.s[0])
                return [NumberAsStringIssue(self.rule, value)]
            except ValueError:
                pass
        return []


@dataclass
class ControlCharacterIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "String contains control character likely caused by missing (double-)slash"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    string: SigmaString


class ControlCharacterValidator(SigmaStringValueValidator):
    """
    Check for control characters in string values, which are normally inserted unintentionally by
    wrong usage of single backslashes, e.g. before a t character, where double backslashes are required.
    """

    def validate_value(self, value: SigmaString) -> List[SigmaValidationIssue]:
        if any((ord(c) < 31 for s in value.s for c in (s if isinstance(s, str) else ""))):
            return [ControlCharacterIssue([self.rule], value)]
        else:
            return []


@dataclass
class WildcardsInsteadOfContainsModifierIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "String contains wildcards at beginning and end instead of being modified with contains modifier"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    detection_item: SigmaDetectionItem


@dataclass
class WildcardInsteadOfStartswithIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "String contains wildcard at end instead of being modified with startswith modifier"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    detection_item: SigmaDetectionItem


@dataclass
class WildcardInsteadOfEndswithIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "String contains wildcard at beginning instead of being modified with endswith modifier"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    detection_item: SigmaDetectionItem


class WildcardsInsteadOfModifiersValidator(SigmaDetectionItemValidator):
    """Check if wildcards were used where usage of startswith, endswith and contains modifiers would be possible."""

    def validate_detection_item(
        self, detection_item: SigmaDetectionItem
    ) -> List[SigmaValidationIssue]:
        # Warning rule use a single '*' waiting for the `exists` modifier  so check len(value)>1 to allow it
        if (
            all(
                (
                    isinstance(value, SigmaString)
                    and len(value) > 1
                    and value.startswith(SpecialChars.WILDCARD_MULTI)
                    and value.endswith(SpecialChars.WILDCARD_MULTI)
                    and not value[1:-1].contains_special()
                    for value in detection_item.original_value
                )
            )
            and SigmaContainsModifier not in detection_item.modifiers
        ):
            return [WildcardsInsteadOfContainsModifierIssue([self.rule], detection_item)]
        elif (
            all(
                (
                    isinstance(value, SigmaString)
                    and len(value) > 1
                    and value.startswith(SpecialChars.WILDCARD_MULTI)
                    and not value[1:].contains_special()
                    for value in detection_item.original_value
                )
            )
            and SigmaEndswithModifier not in detection_item.modifiers
        ):
            return [WildcardInsteadOfEndswithIssue([self.rule], detection_item)]
        elif (
            all(
                (
                    isinstance(value, SigmaString)
                    and len(value) > 1
                    and value.endswith(SpecialChars.WILDCARD_MULTI)
                    and not value[:-1].contains_special()
                    for value in detection_item.original_value
                )
            )
            and SigmaStartswithModifier not in detection_item.modifiers
        ):
            return [WildcardInsteadOfStartswithIssue([self.rule], detection_item)]
        else:
            return []


@dataclass
class EscapedWildcardIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "Rule contains an escaped wildcard in the rule logic. Make sure the escape is intentional."
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW
    string: SigmaString


class EscapedWildcardValidator(SigmaStringValueValidator):
    """Check for the presence of escaped wildcards."""

    wildcard_list = ["*", "?"]

    def validate_value(self, value: SigmaString) -> List[SigmaValidationIssue]:
        if any([x in s for x in self.wildcard_list for s in value if isinstance(s, str)]):
            return [EscapedWildcardIssue(self.rule, value)]
        else:
            return []



from dataclasses import dataclass
from re import Pattern
import re
from typing import ClassVar, List, Set
from sigma.conditions import ConditionIdentifier, ConditionItem, ConditionSelector
from sigma.rule import SigmaDetections, SigmaRule
from sigma.validators.base import (
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
    SigmaRuleValidator,
)


@dataclass
class DanglingDetectionIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule defines detection that is not referenced from condition"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.HIGH
    detection_name: str


class DanglingDetectionValidator(SigmaRuleValidator):
    """Check for detection definitions not referenced from condition."""

    detection_names: Set[str]

    def condition_referenced_ids(
        self, cond: ConditionItem, detections: SigmaDetections
    ) -> Set[str]:
        """
        Return detection item identifierd referenced by condition.

        :param cond: Condition to analyze.
        :type cond: ConditionItem
        :param detections: Detections referenced from condition.
        :type detections: SigmaDetections
        :return: Set of referenced detection identifiers.
        :rtype: Set[str]
        """
        if isinstance(cond, ConditionIdentifier):  # Only one id referenced.
            return {cond.identifier}
        elif isinstance(cond, ConditionSelector):  # Resolve all referenced ids and return
            return {cond.identifier for cond in cond.resolve_referenced_detections(detections)}
        elif isinstance(cond, ConditionItem):  # Traverse into subconditions
            ids = set()
            for arg in cond.args:
                ids.update(self.condition_referenced_ids(arg, detections))
            return ids
        else:  # Fallback if something different is encountered: return empty set.
            return set()

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        detection_names = {  # collect detection names
            name for name in rule.detection.detections.keys()
        }
        referenced_ids = set()
        for condition in rule.detection.parsed_condition:
            parsed_condition = condition.parse(False)
            referenced_ids.update(self.condition_referenced_ids(parsed_condition, rule.detection))

        return [DanglingDetectionIssue([rule], name) for name in detection_names - referenced_ids]


@dataclass
class ThemConditionWithSingleDetectionIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Rule refers to 'them' but has only one condition"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.LOW


class ThemConditionWithSingleDetectionValidator(SigmaRuleValidator):
    """Detect conditions refering to 'them' with only one detection."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if (
            any(["them" in condition for condition in rule.detection.condition])
            and len(rule.detection.detections) == 1
        ):
            return [ThemConditionWithSingleDetectionIssue([rule])]
        else:
            return []


@dataclass
class AllOfThemConditionIssue(SigmaValidationIssue):
    description: ClassVar[
        str
    ] = "Rule contains discouraged 'all of them' condition, use 'all of selection*' instead."
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM


class AllOfThemConditionValidator(SigmaRuleValidator):
    """Find occurrences of discouraged 'all of them' conditions."""

    re_all_of_them: ClassVar[Pattern] = re.compile("all\\s+of\\s+them")

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        if any([self.re_all_of_them.search(condition) for condition in rule.detection.condition]):
            return [AllOfThemConditionIssue([rule])]
        else:
            return []



from dataclasses import InitVar, dataclass, field
import dataclasses
from typing import Any, Dict, Optional, Union, Sequence, List, Mapping, Type
from uuid import UUID
from enum import Enum, auto
from datetime import date, datetime
import yaml
import sigma
from sigma.types import SigmaType, SigmaNull, SigmaString, SigmaNumber, sigma_type
from sigma.modifiers import (
    SigmaModifier,
    modifier_mapping,
    reverse_modifier_mapping,
    SigmaValueModifier,
    SigmaListModifier,
)
from sigma.conditions import (
    SigmaCondition,
    ConditionAND,
    ConditionOR,
    ConditionFieldEqualsValueExpression,
    ConditionValueExpression,
    ParentChainMixin,
)
from sigma.processing.tracking import ProcessingItemTrackingMixin
import sigma.exceptions as sigma_exceptions
from sigma.exceptions import (
    SigmaRuleLocation,
    SigmaTypeError,
    SigmaError,
)


class EnumLowercaseStringMixin:
    def __str__(self) -> str:
        return self.name.lower()


class SigmaStatus(EnumLowercaseStringMixin, Enum):
    STABLE = auto()
    EXPERIMENTAL = auto()
    TEST = auto()
    DEPRECATED = auto()
    UNSUPPORTED = auto()


class SigmaLevel(EnumLowercaseStringMixin, Enum):
    INFORMATIONAL = auto()
    LOW = auto()
    MEDIUM = auto()
    HIGH = auto()
    CRITICAL = auto()


class SigmaRelatedType(EnumLowercaseStringMixin, Enum):
    OBSOLETES = auto()
    RENAMED = auto()
    MERGED = auto()
    DERIVED = auto()
    SIMILAR = auto()


@dataclass
class SigmaRelatedItem:
    id: UUID
    type: SigmaRelatedType

    @classmethod
    def from_dict(cls, value: dict) -> "SigmaRelatedItem":
        """Returns Related item from dict with fields."""
        try:
            id = UUID(value["id"])
        except ValueError:
            raise sigma_exceptions.SigmaRelatedError(f"Sigma related identifier must be an UUID")

        try:
            type = SigmaRelatedType[value["type"].upper()]
        except:
            raise sigma_exceptions.SigmaRelatedError(
                f"{value['type']} is not a Sigma related valid type"
            )

        return cls(
            id,
            type,
        )


@dataclass
class SigmaRelated:
    related: List[Type[SigmaRelatedItem]]

    @classmethod
    def from_dict(cls, val: list) -> "SigmaRelated":
        """Returns Related object from dict with fields."""

        list_ret = []
        for v in val:
            if not "id" in v.keys():
                raise sigma_exceptions.SigmaRelatedError("Sigma related must have an id field")
            elif not "type" in v.keys():
                raise sigma_exceptions.SigmaRelatedError("Sigma related must have a type field")
            else:
                list_ret.append(SigmaRelatedItem.from_dict(v))  # should rise the SigmaRelatedError

        return cls(list_ret)


@dataclass(unsafe_hash=True)
class SigmaRuleTag:
    namespace: str
    name: str
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)

    @classmethod
    def from_str(cls, tag: str, source: Optional[SigmaRuleLocation] = None) -> "SigmaRuleTag":
        """Build SigmaRuleTag class from plain text tag string."""
        try:
            ns, n = tag.split(".", maxsplit=1)
        except ValueError:
            raise sigma_exceptions.SigmaValueError(
                "Sigma tag must start with namespace separated with dot from remaining tag."
            )
        return cls(ns, n)

    def __str__(self) -> str:
        return f"{self.namespace}.{self.name}"


@dataclass(frozen=True)
class SigmaLogSource:
    category: Optional[str] = field(default=None)
    product: Optional[str] = field(default=None)
    service: Optional[str] = field(default=None)
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)

    def __post_init__(self):
        """Ensures that log source is not empty."""
        if self.category is None and self.product is None and self.service is None:
            raise sigma_exceptions.SigmaLogsourceError(
                "Sigma log source can't be empty", source=self.source
            )

    @classmethod
    def from_dict(
        cls, logsource: dict, source: Optional[SigmaRuleLocation] = None
    ) -> "SigmaLogSource":
        """Returns SigmaLogSource object from dict with fields."""
        return cls(
            logsource.get("category"),
            logsource.get("product"),
            logsource.get("service"),
            source,
        )

    def to_dict(self) -> dict:
        return {
            field.name: str(value)
            for field in dataclasses.fields(self)
            if (value := self.__getattribute__(field.name)) is not None
        }

    def __contains__(self, other: "SigmaLogSource") -> bool:
        """
        Matching of log source specifications. A log source contains another one if:

        * Both log sources are equal
        * The log source specifies less attributes than the other and the specified attributes are equal
        """
        if not isinstance(other, self.__class__):
            raise TypeError(
                "Containment check only allowed between log sources", source=self.source
            )

        if self == other:
            return True

        return (
            (self.category is None or self.category == other.category)
            and (self.product is None or self.product == other.product)
            and (self.service is None or self.service == other.service)
        )


@dataclass
class SigmaDetectionItem(ProcessingItemTrackingMixin, ParentChainMixin):
    """
    Single Sigma detection definition

    A detection consists of:
    * an optional field name
    * a list of value modifiers that can also be empty
    * the mandatory value or a list of values (internally it's always a list of values)

    By default all values are OR-linked but the 'all' modifier can be used to override this
    behavior.

    If the `auto_modifiers` parameter is set to False, modifiers are not automatically applied to
    the values. This shouldn't normally be used, but only in test scenarios.
    """

    field: Optional[str]  # if None, this is a keyword argument not bound to a field
    modifiers: List[Type[SigmaModifier]]
    value: List[SigmaType]
    value_linking: Union[Type[ConditionAND], Type[ConditionOR]] = ConditionOR
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)
    original_value: Optional[SigmaType] = field(
        init=False, repr=False, hash=False, compare=False
    )  # Copy of original values for conversion back to data structures (and YAML/JSON)
    auto_modifiers: InitVar[bool] = True

    def __post_init__(self, auto_modifiers):
        if not isinstance(self.value, list):  # value cleanup: it has to be a list!
            self.value = [self.value]
        self.value = [  # value cleanup: convert plain values into SigmaType's
            sigma_type(v) if not isinstance(v, SigmaType) else v for v in self.value
        ]

        self.original_value = self.value.copy()  # Create a copy of original values
        if auto_modifiers:
            self.apply_modifiers()

    def apply_modifiers(self):
        """
        Applies modifiers to detection and values
        """
        applied_modifiers = list()
        for modifier in self.modifiers:
            modifier_instance = modifier(self, applied_modifiers, self.source)
            if isinstance(
                modifier_instance, SigmaValueModifier
            ):  # Value modifiers are applied to each value separately
                self.value = [item for val in self.value for item in modifier_instance.apply(val)]
            elif isinstance(
                modifier_instance, SigmaListModifier
            ):  # List modifiers are applied to the whole value list at once
                self.value = modifier_instance.apply(self.value)
            else:  # pragma: no cover
                raise TypeError(
                    "Instance of SigmaValueModifier or SigmaListModifier was expected",
                    source=self.source,
                )  # This should only happen if wrong mapping is defined, therefore no test for this case
            applied_modifiers.append(modifier)

    @classmethod
    def from_mapping(
        cls,
        key: Optional[str],
        val: Union[
            List[Union[int, float, str]],
            Union[int, float, str],
            None,
        ],
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaDetectionItem":
        """
        Constructs SigmaDetectionItem object from a mapping between field name containing
        modifiers and a value. This also supports keys containing only value modifiers
        which results in a keyword detection.

        The value accepts plain values as well as lists of values and resolves them into
        the value list always contained in a SigmaDetectionItem instance.
        """
        if key is None:  # no key at all means pure keyword detection without value modifiers
            field = None
            modifier_ids = list()
        else:  # key-value detection
            field, *modifier_ids = key.split("|")
            if field == "":
                field = None

        try:
            modifiers = [modifier_mapping[mod_id] for mod_id in modifier_ids]
        except KeyError as e:
            raise sigma_exceptions.SigmaModifierError(f"Unknown modifier {str(e)}", source=source)

        if isinstance(val, (int, float, str)):  # value is plain, convert into single element list
            val = [val]
        elif val is None:
            val = [None]

        # Map Python types to Sigma typing classes
        val = [sigma_type(v) for v in val]

        return cls(field, modifiers, val, source=source)

    @classmethod
    def from_value(
        cls,
        val: Union[
            List[Union[int, str]],
            Union[int, str],
        ],
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaDetectionItem":
        """Convenience method for from_mapping(None, value)."""
        return cls.from_mapping(None, val, source=source)

    def disable_conversion_to_plain(self):
        """
        Mark detection item as not convertible to plain data type. This is required in cases where
        the value and original value get out of sync, e.g. because transformation are applied and
        conversion with to_plain() would yield an outdated state.
        """
        self.original_value = None

    def to_plain(self) -> Union[Dict[str, Union[str, int, None]], List[str]]:
        """
        Convert detection item into plain Python type, that can be:

        * a plain value if it is a single plain keyword value.
        * a list of values if it is a list of keyword values
        * a dict in all other cases (detection item bound to field or keyword with modifiers)
        """
        if self.original_value is None:
            raise sigma_exceptions.SigmaValueError(
                f"Detection item { str(self) } can't be converted to plain data type anymore because the current value is not in sync with original value anymore, e.g. by applying transformations.",
                source=self.source,
            )

        if len(self.original_value) > 1:
            value = [value.to_plain() for value in self.original_value]
        else:
            value = self.original_value[0].to_plain()

        if (
            self.is_keyword() and len(self.modifiers) == 0
        ):  # detection item is keyword detection and has no modifiers: return list of values
            return value
        else:
            field_name = (
                self.field or ""
            )  # field name is empty in case of keyword detection items with modifiers
            modifier_ids = [  # list of modifier identifiers from reverse mapping
                reverse_modifier_mapping[modifier.__name__] for modifier in self.modifiers
            ]
            if len(modifier_ids) > 0:
                modifiers_prefix = "|"
            else:
                modifiers_prefix = ""
            return {field_name + modifiers_prefix + "|".join(modifier_ids): value}

    def postprocess(
        self,
        detections: "SigmaDetections",
        parent: Optional["sigma.condition.ConditionItem"] = None,
    ) -> Union[
        ConditionAND,
        ConditionOR,
        ConditionFieldEqualsValueExpression,
        ConditionValueExpression,
    ]:
        super().postprocess(detections, parent)
        if len(self.value) == 0:  # no value: map to none type
            if self.field is None:
                raise sigma_exceptions.SigmaConditionError(
                    "Null value must be bound to a field", source=self.source
                )
            else:
                return ConditionFieldEqualsValueExpression(self.field, SigmaNull()).postprocess(
                    detections, self, self.source
                )
        if len(self.value) == 1:  # single value: return key/value or value-only expression
            if self.field is None:
                return ConditionValueExpression(self.value[0]).postprocess(
                    detections, self, self.source
                )
            else:
                return ConditionFieldEqualsValueExpression(self.field, self.value[0]).postprocess(
                    detections, self, self.source
                )
        else:  # more than one value, return logically linked values or an "in" expression
            if self.field is None:  # no field - only values
                cond = self.value_linking([ConditionValueExpression(v) for v in self.value])
            else:  # with field - field/value pairs
                cond = self.value_linking(
                    [ConditionFieldEqualsValueExpression(self.field, v) for v in self.value]
                )
            cond.postprocess(detections, parent, self.source)
            return cond

    def is_keyword(self) -> bool:
        """Returns True if detection item is a keyword detection without field reference."""
        return self.field is None


@dataclass
class SigmaDetection(ParentChainMixin):
    """
    A detection is a set of atomic event defitionions represented by SigmaDetectionItem instances. SigmaDetectionItems
    of a SigmaDetection are OR-linked.

    A detection can be defined by:

    1. a mapping between field/value pairs that all should appear in matched events.
    2. a plain value
    3. a list of plain values or mappings defined and matched as in 1 where at least one of the items should appear in matched events.
    """

    detection_items: List[Union[SigmaDetectionItem, "SigmaDetection"]]
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)
    item_linking: Union[Type[ConditionAND], Type[ConditionOR]] = field(default=None)

    def __post_init__(self):
        """Check detection validity."""
        if len(self.detection_items) == 0:
            raise sigma_exceptions.SigmaDetectionError("Detection is empty", source=self.source)

        if self.item_linking is None:
            type_set = {type(item) for item in self.detection_items}
            if SigmaDetectionItem in type_set:
                self.item_linking = ConditionAND
            else:
                self.item_linking = ConditionOR

    @classmethod
    def from_definition(
        cls,
        definition: Union[Mapping, Sequence, str, int],
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaDetection":
        """Instantiate an appropriate SigmaDetection object from a parsed Sigma detection definition."""
        if isinstance(definition, Mapping):  # key-value-definition (case 1)
            return cls(
                detection_items=[
                    SigmaDetectionItem.from_mapping(key, val, source)
                    for key, val in definition.items()
                ],
                source=source,
            )
        elif isinstance(definition, (str, int)):  # plain value (case 2)
            return cls(
                detection_items=[SigmaDetectionItem.from_value(definition, source)],
                source=source,
            )
        elif isinstance(definition, Sequence):  # list of items (case 3)
            if {type(item) for item in definition}.issubset(
                {str, int}
            ):  # list of values: create one detection item containing all values
                return cls(
                    detection_items=[SigmaDetectionItem.from_value(definition, source)],
                    source=source,
                )
            else:
                return cls(
                    detection_items=[
                        SigmaDetection.from_definition(
                            item, source
                        )  # nested SigmaDetection in other cases
                        for item in definition
                    ],
                    source=source,
                )

    def to_plain(self) -> Union[Dict[str, Union[str, int, None]], List[str]]:
        """Returns a dictionary or list representation of the detection."""
        detection_items = [  # first convert all detection items into a Python representation.
            detection_item.to_plain() for detection_item in self.detection_items
        ]
        # Filter out where to_plain() returns None, which causes merging errors.
        # We will have to handle the condition as well in conditions.py
        detection_items = [
            detection_item for detection_item in detection_items if detection_item is not None
        ]
        detection_items_types = {  # create set of types for decision what has to be returned
            type(detection_item) for detection_item in detection_items
        }

        if len(detection_items) == 0:  # pragma: no cover
            return None  # This case is catched by the post init check, so it shouldn't happen.
        if len(detection_items) == 1:  # Only one detection item? Return it as result.
            return detection_items[0]
        else:  # More than one detection iten, it depends now on the types
            if dict in detection_items_types and len(detection_items_types) > 1:
                # Merging dicts with other types isn't possibly, at least not in a simple way.
                # This case can appear in a programmatically instantiated detection, but can't be
                # expressed in a data structure, because there might be only a list or a map.
                # In the future (if there's a need) a possibility would be to collect such items
                # under null or empty keys, but for today I want to keep this simple and simply bail
                # out.
                raise sigma_exceptions.SigmaValueError(
                    "Can't convert detection into plain value because it contains mixed detection item types.",
                    source=self.source,
                )
            elif detection_items_types == {dict}:  # only dict's, merge them together
                merged = dict()
                # Count key appearance for later all modifier addition
                key_count = dict()
                # The following double loop (the second one is no real one, as it operates on a
                # single element dict) merges keys (not fields!) into the merged dict.
                for detection_item_converted, detection_item in zip(
                    detection_items, self.detection_items
                ):
                    for k, v in detection_item_converted.items():
                        if k not in merged:  # key doesn't exists in merged dict: just add
                            merged[k] = v
                        else:  # key collision, now the things get complicated...
                            if "|all" in k:  # key contains 'all' modifier
                                if not isinstance(
                                    merged[k], list
                                ):  # make list from existing all-modified value if it's a plain value
                                    merged[k] = [merged[k]]

                                if isinstance(v, list):  # merging two and-linked lists is possible
                                    merged[k].extend(v)
                                else:
                                    merged[k].append(v)
                            else:  # key collision without all modifier: trying to merge both keys into one and-linked key
                                ev = merged[k]  # already existing value

                                # Value normalization: extract value from signle-valued lists
                                if isinstance(ev, list) and len(ev) == 1:
                                    ev = ev[0]
                                if isinstance(v, list) and len(v) == 1:
                                    v = v[0]

                                # Still lists? Merging lists is not allowed
                                if isinstance(ev, list) or isinstance(v, list):
                                    raise sigma_exceptions.SigmaValueError(
                                        f"Can't merge value lists '{k}' into one item due to different logical linking.",
                                        source=self.source,
                                    )

                                vs = [ev, v]  # The new merged value

                                ak = k + "|all"
                                if (
                                    ak in merged
                                ):  # 'all' key already exist, append to this key if possible
                                    if not isinstance(
                                        merged[ak], list
                                    ):  # ensure that existing 'all' key is a list
                                        merged[ak] = [av]
                                    av = merged[ak]
                                    av.extend(vs)
                                else:  # create new 'all' key from both existing keys
                                    merged[ak] = vs
                                del merged[k]

                return {
                    k: (v[0] if isinstance(v, list) and len(v) == 1 else v)
                    for k, v in merged.items()
                }
            else:  # only lists and plain values, merge them into one list
                merged = list()
                for detection_item_converted in detection_items:
                    if isinstance(
                        detection_item_converted, list
                    ):  # if item is a list, extend result list with it.
                        merged.extend(detection_item_converted)
                    else:
                        merged.append(
                            detection_item_converted
                        )  # if item is a plain value, append it to list
                return merged

    def postprocess(
        self,
        detections: "SigmaDetections",
        parent: Optional["sigma.condition.ConditionItem"] = None,
    ) -> Union[ConditionAND, ConditionOR]:
        """Convert detection item into condition tree element"""
        super().postprocess(detections, parent)
        items = [
            detection_item.postprocess(detections, self) for detection_item in self.detection_items
        ]
        if len(items) == 1:  # no boolean linking required, directly return single element
            return items[0]
        elif len(items) > 1:
            condition = self.item_linking(items)
            condition.postprocess(detections, parent, self.source)
            return condition

    def add_applied_processing_item(
        self, processing_item: Optional["sigma.processing.pipeline.ProcessingItem"]
    ):
        """Propagate processing item to all contained detection items."""
        for detection_item in self.detection_items:
            detection_item.add_applied_processing_item(processing_item)


@dataclass
class SigmaDetections:
    """Sigma detection section including named detections and condition."""

    detections: Dict[str, SigmaDetection]
    condition: List[str]
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)

    def __post_init__(self):
        """Detections sanity checks"""
        if self.detections == dict():
            raise sigma_exceptions.SigmaDetectionError(
                "No detections defined in Sigma rule", source=self.source
            )
        self.parsed_condition = [SigmaCondition(cond, self, self.source) for cond in self.condition]

    @classmethod
    def from_dict(
        cls, detections: dict, source: Optional[SigmaRuleLocation] = None
    ) -> "SigmaDetections":
        try:
            if isinstance(detections["condition"], list):
                condition = detections["condition"]
            else:
                condition = [detections["condition"]]
        except KeyError:
            raise sigma_exceptions.SigmaConditionError(
                "Sigma rule must contain at least one condition", source=source
            )

        return cls(
            detections={
                name: SigmaDetection.from_definition(definition, source)
                for name, definition in detections.items()
                if name not in ("condition",)
            },
            condition=condition,
            source=source,
        )

    def to_dict(self) -> dict:
        detections = {
            identifier: detection.to_plain() for identifier, detection in self.detections.items()
        }
        if len(self.condition) > 1:
            condition = self.condition
        else:
            condition = self.condition[0]

        return {
            **detections,
            "condition": condition,
        }

    def __getitem__(self, key: str) -> SigmaDetection:
        """Get detection by name"""
        return self.detections[key]


class SigmaYAMLLoader(yaml.SafeLoader):
    """Custom YAML loader implementing additional functionality for Sigma."""

    def construct_mapping(self, node, deep=...):
        keys = set()
        for k, v in node.value:
            key = self.construct_object(k, deep=deep)
            if key in keys:
                raise yaml.error.YAMLError("Duplicate key '{k}'")
            else:
                keys.add(key)

        return super().construct_mapping(node, deep)


@dataclass
class SigmaRule(ProcessingItemTrackingMixin):
    """
    A single Sigma rule.
    """

    title: str
    logsource: SigmaLogSource
    detection: SigmaDetections
    id: Optional[UUID] = None
    related: Optional[SigmaRelated] = None
    status: Optional[SigmaStatus] = None
    description: Optional[str] = None
    references: List[str] = field(default_factory=list)
    tags: List[SigmaRuleTag] = field(default_factory=list)
    author: Optional[str] = None
    date: Optional[date] = None
    modified: Optional[date] = None
    fields: List[str] = field(default_factory=list)
    falsepositives: List[str] = field(default_factory=list)
    level: Optional[SigmaLevel] = None

    errors: List[sigma_exceptions.SigmaError] = field(default_factory=list)
    source: Optional[SigmaRuleLocation] = field(default=None, compare=False)
    custom_attributes: Dict[str, Any] = field(compare=False, default_factory=dict)

    def __post_init__(self):
        for field in ("references", "tags", "fields", "falsepositives"):
            if self.__getattribute__(field) is None:
                self.__setattr__(field, [])
        if self.id is not None and not isinstance(
            self.id, UUID
        ):  # Try to convert rule id into UUID object, but keep it if not possible
            try:
                self.id = UUID(self.id)
            except ValueError:
                pass

    @classmethod
    def from_dict(
        cls,
        rule: dict,
        collect_errors: bool = False,
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaRule":
        """
        Convert Sigma rule parsed in dict structure into SigmaRule object.

        if collect_errors is set to False exceptions are collected in the errors property of the resulting
        SigmaRule object. Else the first recognized error is raised as exception.
        """
        errors = []
        # Rule identifier may be empty or must be valid UUID
        rule_id = rule.get("id")
        if rule_id is not None:
            try:
                rule_id = UUID(rule_id)
            except ValueError:
                errors.append(
                    sigma_exceptions.SigmaIdentifierError(
                        "Sigma rule identifier must be an UUID", source=source
                    )
                )

        # Rule related validation
        rule_related = rule.get("related")
        if rule_related is not None:
            if not isinstance(rule_related, list):
                errors.append(
                    sigma_exceptions.SigmaRelatedError(
                        "Sigma rule related must be a list", source=source
                    )
                )
            else:
                try:
                    rule_related = SigmaRelated.from_dict(rule_related)
                except sigma_exceptions.SigmaRelatedError as e:
                    errors.append(e)

        # Rule level validation
        level = rule.get("level")
        if level is not None:
            try:
                level = SigmaLevel[level.upper()]
            except KeyError:
                errors.append(
                    sigma_exceptions.SigmaLevelError(
                        f"'{ level }' is no valid Sigma rule level", source=source
                    )
                )

        # Rule status validation
        status = rule.get("status")
        if status is not None:
            if not isinstance(status, str):
                errors.append(
                    sigma_exceptions.SigmaStatusError(
                        "Sigma rule status cannot be a list", source=source
                    )
                )
            else:
                try:
                    status = SigmaStatus[status.upper()]
                except KeyError:
                    errors.append(
                        sigma_exceptions.SigmaStatusError(
                            f"'{ status }' is no valid Sigma rule status", source=source
                        )
                    )

        # parse rule date if existing
        rule_date = rule.get("date")
        if rule_date is not None:
            if not isinstance(rule_date, date) and not isinstance(rule_date, datetime):
                try:
                    rule_date = date(*(int(i) for i in rule_date.split("/")))
                except ValueError:
                    try:
                        rule_date = date(*(int(i) for i in rule_date.split("-")))
                    except ValueError:
                        errors.append(
                            sigma_exceptions.SigmaDateError(
                                f"Rule date '{ rule_date }' is invalid, must be yyyy/mm/dd or yyyy-mm-dd",
                                source=source,
                            )
                        )

        # parse rule modified if existing
        rule_modified = rule.get("modified")
        if rule_modified is not None:
            if not isinstance(rule_modified, date) and not isinstance(rule_modified, datetime):
                try:
                    rule_modified = date(*(int(i) for i in rule_modified.split("/")))
                except ValueError:
                    try:
                        rule_modified = date(*(int(i) for i in rule_modified.split("-")))
                    except ValueError:
                        errors.append(
                            sigma_exceptions.SigmaModifiedError(
                                f"Rule modified '{ rule_modified }' is invalid, must be yyyy/mm/dd or yyyy-mm-dd",
                                source=source,
                            )
                        )

        # Rule fields validation
        rule_fields = rule.get("fields")
        if rule_fields is not None and not isinstance(rule_fields, list):
            errors.append(
                sigma_exceptions.SigmaFieldsError(
                    "Sigma rule fields must be a list",
                    source=source,
                )
            )

        # Rule falsepositives validation
        rule_falsepositives = rule.get("falsepositives")
        if rule_falsepositives is not None and not isinstance(rule_falsepositives, list):
            errors.append(
                sigma_exceptions.SigmaFalsePositivesError(
                    "Sigma rule falsepositives must be a list",
                    source=source,
                )
            )

        # Rule author validation
        rule_author = rule.get("author")
        if rule_author is not None and not isinstance(rule_author, str):
            errors.append(
                sigma_exceptions.SigmaAuthorError(
                    "Sigma rule author must be a string",
                    source=source,
                )
            )

        # Rule description validation
        rule_description = rule.get("description")
        if rule_description is not None and not isinstance(rule_description, str):
            errors.append(
                sigma_exceptions.SigmaDescriptionError(
                    "Sigma rule description must be a string",
                    source=source,
                )
            )

        # Rule references validation
        rule_references = rule.get("references")
        if rule_references is not None and not isinstance(rule_references, list):
            errors.append(
                sigma_exceptions.SigmaReferencesError(
                    "Sigma rule references must be a list",
                    source=source,
                )
            )

        # Rule title validation
        rule_title = rule.get("title")
        if rule_title is None:
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule must have a title",
                    source=source,
                )
            )
        elif not isinstance(rule_title, str):
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule title must be a string",
                    source=source,
                )
            )
        elif len(rule_title) > 256:
            errors.append(
                sigma_exceptions.SigmaTitleError(
                    "Sigma rule title length must not exceed 256 characters",
                    source=source,
                )
            )

        # parse log source
        logsource = None
        try:
            logsource = SigmaLogSource.from_dict(rule["logsource"], source)
        except KeyError:
            errors.append(
                sigma_exceptions.SigmaLogsourceError(
                    "Sigma rule must have a log source", source=source
                )
            )
        except AttributeError:
            errors.append(
                sigma_exceptions.SigmaLogsourceError(
                    "Sigma logsource must be a valid YAML map", source=source
                )
            )
        except SigmaError as e:
            errors.append(e)

        # parse detections
        detections = None
        try:
            detections = SigmaDetections.from_dict(rule["detection"], source)
        except KeyError:
            errors.append(
                sigma_exceptions.SigmaDetectionError(
                    "Sigma rule must have a detection definitions", source=source
                )
            )
        except SigmaError as e:
            errors.append(e)

        if not collect_errors and errors:
            raise errors[0]

        return cls(
            title=rule_title,
            id=rule_id,
            related=rule_related,
            level=level,
            status=status,
            description=rule_description,
            references=rule_references,
            tags=[SigmaRuleTag.from_str(tag) for tag in rule.get("tags", list())],
            author=rule_author,
            date=rule_date,
            modified=rule_modified,
            logsource=logsource,
            detection=detections,
            fields=rule_fields,
            falsepositives=rule_falsepositives,
            errors=errors,
            source=source,
            custom_attributes={
                k: v
                for k, v in rule.items()
                if k
                not in set(cls.__dataclass_fields__.keys())
                - {"errors", "source", "applied_processing_items"}
            },
        )

    @classmethod
    def from_yaml(cls, rule: str, collect_errors: bool = False) -> "SigmaRule":
        """Convert YAML input string with single document into SigmaRule object."""
        parsed_rule = yaml.load(rule, SigmaYAMLLoader)
        return cls.from_dict(parsed_rule, collect_errors)

    def to_dict(self) -> dict:
        """Convert rule object into dict."""
        d = {
            "title": self.title,
            "logsource": self.logsource.to_dict(),
            "detection": self.detection.to_dict(),
        }
        # Convert to string where possible
        for field in ("id", "status", "level", "author", "description"):
            if (s := self.__getattribute__(field)) is not None:
                d[field] = str(s)

        # copy list of strings
        for field in ("references", "fields", "falsepositives"):
            if len(l := self.__getattribute__(field)) > 0:
                d[field] = l.copy()

        # the special cases
        if len(self.tags) > 0:
            d["tags"] = [str(tag) for tag in self.tags]
        if self.date is not None:
            d["date"] = self.date.isoformat()

        # custom attributes
        d.update(self.custom_attributes)

        return d



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from pyparsing import List


@dataclass
class SigmaRuleLocation:
    """Describes a Sigma source file and optionally a location inside it."""

    path: Path
    line: Optional[int] = None
    char: Optional[int] = None

    def __post_init__(self):
        if isinstance(self.path, str):
            self.path = Path(self.path)

    def __str__(self):
        s = str(self.path.resolve())
        if self.line is not None:
            s += ":" + str(self.line)
            if self.char is not None:
                s += ":" + str(self.char)
        return s


class SigmaError(ValueError):
    """Generic Sigma error and super-class of all Sigma exceptions"""

    def __init__(self, *args, **kwargs):
        try:
            self.source = kwargs["source"]
            del kwargs["source"]
        except KeyError:
            self.source = None
        super().__init__(*args, **kwargs)

    def __str__(self):
        if self.source is not None:
            return super().__str__() + " in " + str(self.source)
        else:
            return super().__str__()

    def __eq__(self, other: object) -> bool:
        try:
            return (
                type(self) is type(other)
                and self.source == other.source
                and self.args == other.args
            )
        except AttributeError:
            return False


class SigmaTitleError(SigmaError):
    """Error in Sigma rule logosurce specification"""

    pass


class SigmaLogsourceError(SigmaError):
    """Error in Sigma rule logosurce specification"""

    pass


class SigmaDetectionError(SigmaError):
    """Error in Sigma rule detection"""

    pass


class SigmaConditionError(SigmaError):
    """Error in Sigma rule condition"""

    pass


class SigmaIdentifierError(SigmaError):
    """Error in Sigma rule identifier"""

    pass


class SigmaAuthorError(SigmaError):
    """Error in Sigma rule author"""

    pass


class SigmaRelatedError(SigmaError):
    """Error in Sigma rule related"""

    pass


class SigmaDateError(SigmaError):
    """Error in Sigma rule date"""

    pass


class SigmaModifiedError(SigmaError):
    """Error in Sigma rule modified"""

    pass


class SigmaDescriptionError(SigmaError):
    """Error in Sigma rule description"""

    pass


class SigmaReferencesError(SigmaError):
    """Error in Sigma rule references"""

    pass


class SigmaFieldsError(SigmaError):
    """Error in Sigma rule fields"""

    pass


class SigmaFalsePositivesError(SigmaError):
    """Error in Sigma rule falsepositives"""

    pass


class SigmaStatusError(SigmaError):
    """Error in Sigma rule status"""

    pass


class SigmaLevelError(SigmaError):
    """Error in Sigma rule level"""

    pass


class SigmaModifierError(SigmaError):
    """Error in Sigma rule value modifier specification"""

    pass


class SigmaTypeError(SigmaModifierError):
    """Sigma modifier not applicable on value type"""

    pass


class SigmaValueError(SigmaError):
    """Error in Sigma rule value"""

    pass


class SigmaRegularExpressionError(SigmaValueError):
    """Error in regular expression contained in Sigma rule"""

    pass


class SigmaPlaceholderError(SigmaValueError):
    """Attempted to convert an unhandled Placeholder into a query"""

    pass


class SigmaCollectionError(SigmaError):
    """Error in Sigma collection, e.g. unknown action"""

    pass


class SigmaConfigurationError(SigmaError):
    """Error in configuration of a Sigma processing pipeline"""

    pass


class SigmaFeatureNotSupportedByBackendError(SigmaError):
    """Sigma feature is not supported by the backend."""

    pass


class SigmaPipelineNotFoundError(SigmaError, ValueError):
    """An attempt to resolve a processing pipeline from a specifier failed because it was not
    found."""

    def __init__(self, spec: str, *args, **kwargs):
        self.spec = spec
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"Processing pipeline '{self.spec}' not found"


class SigmaPipelineNotAllowedForBackendError(SigmaConfigurationError):
    """One or multiple processing pipelines doesn't matches the given backend."""

    def __init__(self, spec: str, backend: str, *args, **kwargs):
        self.wrong_pipeline = spec
        self.backend = backend
        super().__init__(*args, **kwargs)

    def __str__(self):
        return (
            f"Processing pipelines not allowed for backend '{self.backend}': {self.wrong_pipeline}"
        )


class SigmaTransformationError(SigmaError):
    """Error while transformation. Can be raised intentionally by FailureTransformation."""


class SigmaPluginNotFoundError(SigmaError):
    """Plugin was not found."""



from abc import abstractmethod
from dataclasses import dataclass
import json
from typing import Any, Dict, List, Literal, Optional

import yaml
import sigma
from sigma.exceptions import SigmaConfigurationError

from sigma.processing.templates import TemplateBase


@dataclass
class Finalizer:
    """Conversion output transformation base class."""

    @classmethod
    def from_dict(cls, d: dict) -> "Finalizer":
        try:
            return cls(**d)
        except TypeError as e:
            raise SigmaConfigurationError("Error in instantiation of finalizer: " + str(e))

    @abstractmethod
    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", queries: List[Any]
    ) -> Any:
        """Finalize output by applying a transformation to the list of generated and postprocessed queries.

        :param pipeline: Processing pipeline this transformation was contained.
        :type pipeline: sigma.processing.pipeline.ProcessingPipeline
        :param queries: List of converted and postprocessed queries that should be finalized.
        :type queries: List[Any]
        :return: Output that can be used in further processing of the conversion result.
        :rtype: Any
        """


@dataclass
class ConcatenateQueriesFinalizer(Finalizer):
    """Concatenate queries with a given separator and embed result within a prefix or suffix
    string."""

    separator: str = "\n"
    prefix: str = ""
    suffix: str = ""

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", queries: List[str]
    ) -> str:
        return self.prefix + self.separator.join(queries) + self.suffix


@dataclass
class JSONFinalizer(Finalizer):
    indent: Optional[int] = None

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", queries: List[Any]
    ) -> str:
        return json.dumps(queries, indent=self.indent)


@dataclass
class YAMLFinalizer(Finalizer):
    indent: Optional[int] = None

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", queries: List[Any]
    ) -> str:
        yaml.safe_dump(queries, indent=self.indent)


@dataclass
class TemplateFinalizer(Finalizer, TemplateBase):
    """Apply Jinja2 template provided as template object variable to the queries. The following
    variables are available in the context:

    * queries: all post-processed queries generated by the backend.
    * pipeline: the Sigma processing pipeline where this transformation is applied including all
      current state information in pipeline.state.

    if *path* is given, *template* is considered as a relative path to a template file below the
    specified path. If it is not provided, the template is specified as plain string. *autoescape*
    controls the Jinja2 HTML/XML auto-escaping.
    """

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", queries: List[Any]
    ) -> str:
        return self.j2template.render(queries=queries, pipeline=pipeline)


finalizers: Dict[str, Finalizer] = {
    "concat": ConcatenateQueriesFinalizer,
    "json": JSONFinalizer,
    "yaml": YAMLFinalizer,
    "template": TemplateFinalizer,
}



from collections import defaultdict
from dataclasses import dataclass, field
from functools import partial
from typing import (
    FrozenSet,
    List,
    Literal,
    Mapping,
    Set,
    Any,
    Callable,
    Iterable,
    Dict,
    Tuple,
    Optional,
    Type,
    Union,
)
from sigma.processing.finalization import Finalizer, finalizers
from sigma.processing.postprocessing import QueryPostprocessingTransformation
from sigma.processing.tracking import FieldMappingTracking
from sigma.rule import SigmaDetectionItem, SigmaRule
from sigma.processing.transformations import transformations, Transformation
from sigma.processing.postprocessing import query_postprocessing_transformations
from sigma.processing.conditions import (
    rule_conditions,
    RuleProcessingCondition,
    detection_item_conditions,
    DetectionItemProcessingCondition,
    field_name_conditions,
    FieldNameProcessingCondition,
)
from sigma.exceptions import SigmaConfigurationError, SigmaTypeError
import yaml

from sigma.types import SigmaFieldReference, SigmaType


@dataclass
class ProcessingItemBase:
    transformation: Transformation
    rule_condition_linking: Callable[[Iterable[bool]], bool] = all  # any or all
    rule_condition_negation: bool = False
    rule_conditions: List[RuleProcessingCondition] = field(default_factory=list)
    identifier: Optional[str] = None

    @classmethod
    def _parse_conditions(
        self,
        condition_class_mapping: Dict[
            str,
            Union[
                Type[RuleProcessingCondition],
                Type[DetectionItemProcessingCondition],
                Type[FieldNameProcessingCondition],
            ],
        ],
        cond_defs: Dict,
    ) -> Union[
        List[RuleProcessingCondition],
        List[DetectionItemProcessingCondition],
        List[FieldNameProcessingCondition],
    ]:
        """Parse dict of conditions into list of condition object instances.

        :param condition_class_mapping: Mapping between condition type identifiers and condition classes.
        :type condition_class_mapping: Dict[str, Union[Type[RuleProcessingCondition], Type[DetectionItemProcessingCondition], Type[FieldNameProcessingCondition]]]
        :param cond_defs: Definition of conditions for the pipeline.
        :type cond_defs: Dict
        :return: List of condition classes as defined in dict.
        :rtype: Union[List[RuleProcessingCondition], List[DetectionItemProcessingCondition], List[FieldNameProcessingCondition]]
        """
        conds = list()
        for i, cond_def in enumerate(cond_defs):
            try:
                cond_type = cond_def["type"]
            except KeyError:
                raise SigmaConfigurationError(
                    f"Missing condition type defined in condition { i + 1 }"
                )

            try:
                cond_class = condition_class_mapping[cond_type]
            except KeyError:
                raise SigmaConfigurationError(
                    f"Unknown condition type '{ cond_type }' in condition { i + 1 }"
                )

            cond_params = {k: v for k, v in cond_def.items() if k != "type"}
            try:
                conds.append(cond_class(**cond_params))
            except (SigmaConfigurationError, TypeError) as e:
                raise SigmaConfigurationError(f"Error in condition { i + 1 }: { str(e) }") from e
        return conds

    @classmethod
    def _instantiate_transformation(cls, d: dict, transformations: Dict[str, Type[Transformation]]):
        try:
            transformation_class_name = d["type"]
        except KeyError:
            raise SigmaConfigurationError("Missing transformation type")

        try:
            return transformations[transformation_class_name]
        except KeyError:
            raise SigmaConfigurationError(
                f"Unknown transformation type '{ transformation_class_name }'"
            )

    def match_rule_conditions(self, pipeline: "ProcessingPipeline", rule: SigmaRule):
        cond_result = self.rule_condition_linking(
            [condition.match(pipeline, rule) for condition in self.rule_conditions]
        )
        if self.rule_condition_negation:
            cond_result = not cond_result
        return not self.rule_conditions or cond_result


@dataclass
class ProcessingItem(ProcessingItemBase):
    """
    A processing item consists of an optional condition and a transformation that is applied in the case that
    the condition evaluates to true against the given Sigma rule or if the condition is not present.

    Processing items are instantiated by the processing pipeline for a whole collection that is about to be
    converted by a backend.
    """

    detection_item_condition_linking: Callable[[Iterable[bool]], bool] = all  # any or all
    detection_item_condition_negation: bool = False
    detection_item_conditions: List[DetectionItemProcessingCondition] = field(default_factory=list)
    field_name_condition_linking: Callable[[Iterable[bool]], bool] = all  # any or all
    field_name_condition_negation: bool = False
    field_name_conditions: List[FieldNameProcessingCondition] = field(default_factory=list)

    @classmethod
    def from_dict(cls, d: dict):
        """Instantiate processing item from parsed definition and variables."""
        # Identifier
        identifier = d.get("id", None)

        # Parse conditions
        rule_conds = cls._parse_conditions(rule_conditions, d.get("rule_conditions", list()))
        detection_item_conds = cls._parse_conditions(
            detection_item_conditions, d.get("detection_item_conditions", list())
        )
        field_name_conds = cls._parse_conditions(
            field_name_conditions, d.get("field_name_conditions", list())
        )

        condition_linking = {
            "or": any,
            "and": all,
        }
        rule_condition_linking = condition_linking[
            d.get("rule_cond_op", "and")
        ]  # default: conditions are linked with and operator
        detection_item_condition_linking = condition_linking[
            d.get("detection_item_cond_op", "and")
        ]  # same for detection item conditions
        field_name_condition_linking = condition_linking[
            d.get("field_name_cond_op", "and")
        ]  # same for field name conditions

        rule_condition_negation = d.get("rule_cond_not", False)
        detection_item_condition_negation = d.get("detection_item_cond_not", False)
        field_name_condition_negation = d.get("field_name_cond_not", False)

        # Transformation
        transformation_class = cls._instantiate_transformation(d, transformations)

        params = {
            k: v
            for k, v in d.items()
            if k
            not in {
                "rule_conditions",
                "rule_cond_op",
                "rule_cond_not",
                "detection_item_conditions",
                "detection_item_cond_op",
                "detection_item_cond_not",
                "field_name_conditions",
                "field_name_cond_op",
                "field_name_cond_not",
                "type",
                "id",
            }
        }
        try:
            transformation = transformation_class(**params)
        except (SigmaConfigurationError, TypeError) as e:
            raise SigmaConfigurationError("Error in transformation: " + str(e)) from e

        return cls(
            transformation,
            rule_condition_linking,
            rule_condition_negation,
            rule_conds,
            identifier,
            detection_item_condition_linking,
            detection_item_condition_negation,
            detection_item_conds,
            field_name_condition_linking,
            field_name_condition_negation,
            field_name_conds,
        )

    def __post_init__(self):
        self.transformation.set_processing_item(
            self
        )  # set processing item in transformation object after it is instantiated
        if not isinstance(self.rule_conditions, list):
            raise SigmaTypeError("Rule processing conditions must be provided as list")
        for rule_condition in self.rule_conditions:
            if not isinstance(rule_condition, RuleProcessingCondition):
                raise SigmaTypeError(
                    f"Rule processing condition '{str(rule_condition)}' is not a RuleProcessingCondition"
                )
        if not isinstance(self.detection_item_conditions, list):
            raise SigmaTypeError("Detection item processing conditions must be provided as list")
        for detection_item_condition in self.detection_item_conditions:
            if not isinstance(detection_item_condition, DetectionItemProcessingCondition):
                raise SigmaTypeError(
                    f"Detection item processing condition '{str(detection_item_condition)}' is not a DetectionItemProcessingCondition"
                )
        if not isinstance(self.field_name_conditions, list):
            raise SigmaTypeError("Field name processing conditions must be provided as list")
        for field_name_condition in self.field_name_conditions:
            if not isinstance(field_name_condition, FieldNameProcessingCondition):
                raise SigmaTypeError(
                    f"Detection item processing condition '{str(field_name_condition)}' is not a FieldNameProcessingCondition"
                )

    def apply(self, pipeline: "ProcessingPipeline", rule: SigmaRule) -> bool:
        """
        Matches condition against rule and performs transformation if condition is true or not present.
        Returns Sigma rule and bool if transformation was applied.
        """
        if self.match_rule_conditions(
            pipeline, rule
        ):  # apply transformation if conditions match or no condition defined
            self.transformation.apply(pipeline, rule)
            return True
        else:  # just pass rule through
            return False

    def match_detection_item(
        self, pipeline: "ProcessingPipeline", detection_item: SigmaDetectionItem
    ) -> bool:
        """
        Evalutates detection item and field name conditions from processing item to detection item
        and returns result.
        """
        detection_item_cond_result = self.detection_item_condition_linking(
            [
                condition.match(pipeline, detection_item)
                for condition in self.detection_item_conditions
            ]
        )
        if self.detection_item_condition_negation:
            detection_item_cond_result = not detection_item_cond_result

        field_name_cond_result = self.field_name_condition_linking(
            [
                condition.match_detection_item(pipeline, detection_item)
                for condition in self.field_name_conditions
            ]
        )
        if self.field_name_condition_negation:
            field_name_cond_result = not field_name_cond_result

        return detection_item_cond_result and field_name_cond_result

    def match_field_name(self, pipeline: "ProcessingPipeline", field: Optional[str]) -> bool:
        """
        Evaluate field name conditions on field names and return result.
        """
        field_name_cond_result = self.field_name_condition_linking(
            [
                condition.match_field_name(pipeline, field)
                for condition in self.field_name_conditions
            ]
        )
        if self.field_name_condition_negation:
            field_name_cond_result = not field_name_cond_result

        return field_name_cond_result

    def match_field_in_value(self, pipeline: "ProcessingPipeline", value: SigmaType) -> bool:
        """
        Evaluate field name conditions in field reference values and return result.
        """
        if isinstance(value, SigmaFieldReference):
            field_name_cond_result = self.field_name_condition_linking(
                [condition.match_value(pipeline, value) for condition in self.field_name_conditions]
            )
            if self.field_name_condition_negation:
                field_name_cond_result = not field_name_cond_result

            return field_name_cond_result
        else:
            return False


@dataclass
class QueryPostprocessingItem(ProcessingItemBase):
    """
    A query post-processing item consists of an optional rule condition and a post-processing
    transformation that operates on the queries that were emitted by a backend for a given rule
    in the case that a transformation that is applied the condition evaluates to true against the
    given Sigma rule or if the condition is not present.

    These items are instantiated by the processing pipeline for a whole collection that was
    converted by a backend.
    """

    @classmethod
    def from_dict(cls, d: dict):
        """Instantiate processing item from parsed definition and variables."""
        # Identifier
        identifier = d.get("id", None)

        # Parse conditions
        condition_linking = {
            "or": any,
            "and": all,
        }
        rule_conds = cls._parse_conditions(rule_conditions, d.get("rule_conditions", list()))
        rule_condition_linking = condition_linking[
            d.get("rule_cond_op", "and")
        ]  # default: conditions are linked with and operator
        rule_condition_negation = d.get("rule_cond_not", False)

        # Transformation
        transformation_class = cls._instantiate_transformation(
            d, query_postprocessing_transformations
        )

        params = {
            k: v
            for k, v in d.items()
            if k not in {"rule_conditions", "rule_cond_op", "rule_cond_not", "type", "id"}
        }
        try:
            transformation = transformation_class(**params)
        except (SigmaConfigurationError, TypeError) as e:
            raise SigmaConfigurationError("Error in transformation: " + str(e)) from e

        return cls(
            transformation, rule_condition_linking, rule_condition_negation, rule_conds, identifier
        )

    def __post_init__(self):
        self.transformation.set_processing_item(
            self
        )  # set processing item in transformation object after it is instantiated
        if not isinstance(self.rule_conditions, list):
            raise SigmaTypeError("Rule processing conditions must be provided as list")
        for rule_condition in self.rule_conditions:
            if not isinstance(rule_condition, RuleProcessingCondition):
                raise SigmaTypeError(
                    f"Rule processing condition '{str(rule_condition)}' is not a RuleProcessingCondition"
                )

    def apply(
        self, pipeline: "ProcessingPipeline", rule: SigmaRule, query: str
    ) -> Tuple[str, bool]:
        """
        Matches condition against rule and performs transformation of query if condition is true or not present.
        Returns query and bool if transformation was applied.
        """
        if self.match_rule_conditions(
            pipeline, rule
        ):  # apply transformation if conditions match or no condition defined
            result = self.transformation.apply(pipeline, rule, query)
            return (result, True)
        else:  # just pass rule through
            return (query, False)


@dataclass
class ProcessingPipeline:
    """
    A processing pipeline is configured with the transformation steps that are applied on Sigma rules and
    are configured by:

    * a backend to apply a set of base preprocessing of Sigma rules (e.g. renaming of fields).
    * the user in one or multiple configurations to conduct further rule transformation to adapt the rule
      to the environment.

    A processing pipeline is instantiated once for a rule collection. Rules are processed in order of their
    appearance in a rule file or include order. Further, processing pipelines can be chained and contain
    variables that can be used from processing items.
    """

    items: List[ProcessingItem] = field(default_factory=list)
    postprocessing_items: List[QueryPostprocessingTransformation] = field(default_factory=list)
    finalizers: List[Finalizer] = field(default_factory=list)
    vars: Dict[str, Any] = field(default_factory=dict)
    priority: int = field(default=0)
    name: Optional[str] = field(default=None)
    allowed_backends: FrozenSet[str] = field(
        default_factory=frozenset
    )  # Set of identifiers of backends (from the backends mapping) that are allowed to use this processing pipeline. This can be used by frontends like Sigma CLI to warn the user about inappropriate usage.
    # The following items are reset for each invocation of apply().
    # TODO: move this to parameters or return values of apply().
    applied: List[bool] = field(
        init=False, compare=False, default_factory=list
    )  # list of applied items as booleans. If True, the corresponding item at the same position was applied
    applied_ids: Set[str] = field(
        init=False, compare=False, default_factory=set
    )  # set of identifiers of applied items, doesn't contains items without identifier
    field_name_applied_ids: Dict[str, Set[str]] = field(
        init=False, compare=False, default_factory=partial(defaultdict, set)
    )  # Mapping of field names from rule fields list to set of applied processing items
    field_mappings: FieldMappingTracking = field(
        init=False, compare=False, default_factory=FieldMappingTracking
    )  # Mapping between initial field names and finally mapped field name.
    state: Mapping[str, Any] = field(
        init=False, compare=False, default_factory=dict
    )  # pipeline state: allows to set variables that can be used in conversion (e.g. indices, data model names etc.)

    def __post_init__(self):
        if not all((isinstance(item, ProcessingItem) for item in self.items)):
            raise TypeError(
                "Each item in a processing pipeline must be a ProcessingItem - don't use processing classes directly!"
            )

    @classmethod
    def from_dict(cls, d: dict) -> "ProcessingPipeline":
        """Instantiate processing pipeline from a parsed processing item description."""
        vars = d.get("vars", dict())  # default: no variables

        items = d.get("transformations", list())  # default: no transformation
        processing_items = list()
        for i, item in enumerate(items):
            try:
                processing_items.append(ProcessingItem.from_dict(item))
            except SigmaConfigurationError as e:
                raise SigmaConfigurationError(
                    f"Error in processing rule { i + 1 }: { str(e) }"
                ) from e

        items = d.get("postprocessing", list())  # default: no transformation
        postprocessing_items = list()
        for i, item in enumerate(items):
            try:
                postprocessing_items.append(QueryPostprocessingItem.from_dict(item))
            except SigmaConfigurationError as e:
                raise SigmaConfigurationError(
                    f"Error in processing rule { i + 1 }: { str(e) }"
                ) from e

        fds = d.get("finalizers", list())  # no default transformation
        fs = list()
        for fd in fds:
            try:
                finalizer_type = fd.pop("type")
            except KeyError:
                raise SigmaConfigurationError(
                    "Finalizer type must be specified in 'type' attribute"
                )

            try:
                fs.append(finalizers[finalizer_type].from_dict(fd))
            except KeyError:
                raise SigmaConfigurationError(f"Finalizer '{finalizer_type}' is unknown")

        priority = d.get("priority", 0)
        name = d.get("name", None)
        allowed_backends = frozenset(d.get("allowed_backends", frozenset()))

        return cls(
            processing_items,
            postprocessing_items,
            fs,
            vars,
            priority,
            name,
            allowed_backends,
        )

    @classmethod
    def from_yaml(cls, processing_pipeline: str) -> "ProcessingPipeline":
        """Convert YAML input string into processing pipeline."""
        parsed_pipeline = yaml.safe_load(processing_pipeline)
        return cls.from_dict(parsed_pipeline)

    def apply(self, rule: SigmaRule) -> SigmaRule:
        """Apply processing pipeline on Sigma rule."""
        self.applied = list()
        self.applied_ids = set()
        self.field_name_applied_ids = defaultdict(set)
        self.field_mappings = FieldMappingTracking()
        self.state = dict()
        for item in self.items:
            applied = item.apply(self, rule)
            self.applied.append(applied)
            if applied and (itid := item.identifier):
                self.applied_ids.add(itid)
        return rule

    def postprocess_query(self, rule: SigmaRule, query: Any) -> Any:
        """Post-process queries with postprocessing_items."""
        for item in self.postprocessing_items:
            query, applied = item.apply(self, rule, query)
            if applied and (itid := item.identifier):
                self.applied_ids.add(itid)
        return query

    def finalize(self, output: Any) -> Any:
        for finalizer in self.finalizers:
            output = finalizer.apply(self, output)
        return output

    def track_field_processing_items(
        self, src_field: str, dest_field: List[str], processing_item_id: Optional[str]
    ) -> None:
        """
        Track processing items that were applied to field names. This adds the processing_item_id to
        the set of applied processing items from src_field and assigns a copy of this set ass
        tracking set to all fields in dest_field.
        """
        if [src_field] != dest_field:  # Only add if source field was mapped to something different.
            applied_identifiers: Set = self.field_name_applied_ids[src_field]
            if processing_item_id is not None:
                applied_identifiers.add(processing_item_id)
            del self.field_name_applied_ids[src_field]
            for field in dest_field:
                self.field_name_applied_ids[field] = applied_identifiers.copy()

    def field_was_processed_by(self, field: Optional[str], processing_item_id: str) -> bool:
        """
        Check if field name was processed by a particular processing item.
        """
        if field is None:
            return False
        return processing_item_id in self.field_name_applied_ids[field]

    def __add__(self, other: Optional["ProcessingPipeline"]) -> "ProcessingPipeline":
        """Concatenate two processing pipelines and merge their variables."""
        if other is None:
            return self
        if not isinstance(other, self.__class__):
            raise TypeError("Processing pipeline must be merged with another one.")

        return self.__class__(
            items=self.items + other.items,
            postprocessing_items=self.postprocessing_items + other.postprocessing_items,
            finalizers=self.finalizers + other.finalizers,
            vars={**self.vars, **other.vars},
        )

    def __radd__(self, other: Literal[0]) -> "ProcessingPipeline":
        """Ignore integer 0 on addition to make sum of list of ProcessingPipelines working."""
        if other == 0:
            return self
        else:
            return NotImplemented



from typing import List
from sigma.processing.conditions import (
    LogsourceCondition,
    RuleContainsDetectionItemCondition,
)
from sigma.processing.pipeline import ProcessingItem
from sigma.processing.transformations import AddConditionTransformation

windows_logsource_mapping = {  # Mapping between Sigma log source services and Windows event log channel names
    "security": "Security",
    "application": "Application",
    "system": "System",
    "sysmon": "Microsoft-Windows-Sysmon/Operational",
    "powershell": [
        "Microsoft-Windows-PowerShell/Operational",
        "PowerShellCore/Operational",
    ],
    "powershell-classic": "Windows PowerShell",
    "taskscheduler": "Microsoft-Windows-TaskScheduler/Operational",
    "wmi": "Microsoft-Windows-WMI-Activity/Operational",
    "dns-server": "DNS Server",
    "dns-server-audit": "Microsoft-Windows-DNS-Server/Audit",
    "dns-server-analytic": "Microsoft-Windows-DNS-Server/Analytical",
    "driver-framework": "Microsoft-Windows-DriverFrameworks-UserMode/Operational",
    "ntlm": "Microsoft-Windows-NTLM/Operational",
    "dhcp": "Microsoft-Windows-DHCP-Server/Operational",
    "msexchange-management": "MSExchange Management",
    "applocker": [
        "Microsoft-Windows-AppLocker/MSI and Script",
        "Microsoft-Windows-AppLocker/EXE and DLL",
        "Microsoft-Windows-AppLocker/Packaged app-Deployment",
        "Microsoft-Windows-AppLocker/Packaged app-Execution",
    ],
    "printservice-admin": "Microsoft-Windows-PrintService/Admin",
    "printservice-operational": "Microsoft-Windows-PrintService/Operational",
    "codeintegrity-operational": "Microsoft-Windows-CodeIntegrity/Operational",
    "smbclient-security": "Microsoft-Windows-SmbClient/Security",
    "firewall-as": "Microsoft-Windows-Windows Firewall With Advanced Security/Firewall",
    "bits-client": "Microsoft-Windows-Bits-Client/Operational",
    "windefend": "Microsoft-Windows-Windows Defender/Operational",
    "terminalservices-localsessionmanager": "Microsoft-Windows-TerminalServices-LocalSessionManager/Operational",
    "microsoft-servicebus-client": "Microsoft-ServiceBus-Client",
    "ldap_debug": "Microsoft-Windows-LDAP-Client/Debug",
    "security-mitigations": [
        "Microsoft-Windows-Security-Mitigations/Kernel Mode",
        "Microsoft-Windows-Security-Mitigations/User Mode",
    ],
    "diagnosis-scripted": "Microsoft-Windows-Diagnosis-Scripted/Operational",
    "shell-core": "Microsoft-Windows-Shell-Core/Operational",
    "openssh": "OpenSSH/Operational",
    "bitlocker": "Microsoft-Windows-BitLocker/BitLocker Management",
    "vhdmp": "Microsoft-Windows-VHDMP/Operational",
    "appxdeployment-server": "Microsoft-Windows-AppXDeploymentServer/Operational",
    "lsa-server": "Microsoft-Windows-LSA/Operational",
    "appxpackaging-om": "Microsoft-Windows-AppxPackaging/Operational",
    "dns-client": "Microsoft-Windows-DNS Client Events/Operational",
    "appmodel-runtime": "Microsoft-Windows-AppModel-Runtime/Admin",
    "capi2": "Microsoft-Windows-CAPI2/Operational",
    "certificateservicesclient-lifecycle-system": "Microsoft-Windows-CertificateServicesClient-Lifecycle-System/Operational",
}


def logsource_windows(service: str) -> LogsourceCondition:
    return LogsourceCondition(
        product="windows",
        service=service,
    )


def logsource_linux(service: str) -> LogsourceCondition:
    return LogsourceCondition(
        product="linux",
        service=service,
    )


def logsource_macos(service: str) -> LogsourceCondition:
    return LogsourceCondition(
        product="macos",
        service=service,
    )


def logsource_category(category: str) -> LogsourceCondition:
    return LogsourceCondition(
        category=category,
    )


def logsource_windows_process_creation() -> LogsourceCondition:
    return LogsourceCondition(
        category="process_creation",
        product="windows",
    )


def logsource_windows_registry_add() -> LogsourceCondition:
    return LogsourceCondition(
        category="registry_add",
        product="windows",
    )


def logsource_windows_registry_set() -> LogsourceCondition:
    return LogsourceCondition(
        category="registry_set",
        product="windows",
    )


def logsource_windows_registry_delete() -> LogsourceCondition:
    return LogsourceCondition(
        category="registry_delete",
        product="windows",
    )


def logsource_windows_registry_event() -> LogsourceCondition:
    return LogsourceCondition(
        category="registry_event",
        product="windows",
    )


def logsource_windows_file_change() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_change",
        product="windows",
    )


def logsource_windows_file_event() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_event",
        product="windows",
    )


def logsource_windows_file_delete() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_delete",
        product="windows",
    )


def logsource_windows_file_access() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_access",
        product="windows",
    )


def logsource_windows_file_rename() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_rename",
        product="windows",
    )


def logsource_windows_image_load() -> LogsourceCondition:
    return LogsourceCondition(
        category="image_load",
        product="windows",
    )


def logsource_windows_pipe_created() -> LogsourceCondition:
    return LogsourceCondition(
        category="pipe_created",
        product="windows",
    )


def logsource_windows_ps_classic_start() -> LogsourceCondition:
    return LogsourceCondition(
        category="ps_classic_start",
        product="windows",
    )


def logsource_windows_ps_module() -> LogsourceCondition:
    return LogsourceCondition(
        category="ps_module",
        product="windows",
    )


def logsource_windows_ps_script() -> LogsourceCondition:
    return LogsourceCondition(
        category="ps_script",
        product="windows",
    )


def logsource_windows_process_access() -> LogsourceCondition:
    return LogsourceCondition(
        category="process_access",
        product="windows",
    )


def logsource_windows_raw_access_thread() -> LogsourceCondition:
    return LogsourceCondition(
        category="raw_access_thread",
        product="windows",
    )


def logsource_windows_wmi_event() -> LogsourceCondition:
    return LogsourceCondition(
        category="wmi_event",
        product="windows",
    )


def logsource_windows_driver_load() -> LogsourceCondition:
    return LogsourceCondition(
        category="driver_load",
        product="windows",
    )


def logsource_windows_create_stream_hash() -> LogsourceCondition:
    return LogsourceCondition(
        category="create_stream_hash",
        product="windows",
    )


def logsource_windows_create_remote_thread() -> LogsourceCondition:
    return LogsourceCondition(
        category="create_remote_thread",
        product="windows",
    )


def logsource_windows_network_connection() -> LogsourceCondition:
    return LogsourceCondition(
        category="network_connection",
        product="windows",
    )


def logsource_windows_network_connection_initiated(
    initiated: bool,
) -> RuleContainsDetectionItemCondition:
    return RuleContainsDetectionItemCondition(
        field="Initiated",
        value="true" if initiated else "false",
    )


def logsource_windows_dns_query() -> LogsourceCondition:
    return LogsourceCondition(
        category="dns_query",
        product="windows",
    )


def logsource_linux_process_creation() -> LogsourceCondition:
    return LogsourceCondition(
        category="process_creation",
        product="linux",
    )


def logsource_linux_network_connection() -> LogsourceCondition:
    return LogsourceCondition(
        category="network_connection",
        product="linux",
    )


def logsource_linux_file_create() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_create",
        product="linux",
    )


def logsource_macos_process_creation() -> LogsourceCondition:
    return LogsourceCondition(
        category="process_creation",
        product="macos",
    )


def logsource_macos_file_create() -> LogsourceCondition:
    return LogsourceCondition(
        category="file_create",
        product="macos",
    )


def logsource_azure_riskdetection() -> LogsourceCondition:
    return LogsourceCondition(
        category="riskdetection",
        product="azure",
    )


def logsource_azure_pim() -> LogsourceCondition:
    return LogsourceCondition(
        category="pim",
        product="azure",
    )


def logsource_azure_auditlogs() -> LogsourceCondition:
    return LogsourceCondition(
        category="auditlogs",
        product="azure",
    )


def logsource_azure_azureactivity() -> LogsourceCondition:
    return LogsourceCondition(
        category="azureactivity",
        product="azure",
    )


def logsource_azure_signinlogs() -> LogsourceCondition:
    return LogsourceCondition(
        category="signinlogs",
        product="azure",
    )


def generate_windows_logsource_items(
    cond_field_template: str,
    cond_value_template: str,
    identifier_template: str = "windows_logsource_{service}",
) -> List[ProcessingItem]:
    """Generate processing items for all Windows logsource mappings from templates. All templates
    are defined as Python f-string ("{variable}"). Available variables in each template are:

    * service: Sigma log source definition field 'service'. Example: security
    * source: Windows log source name. Example: Microsoft-Windows-Sysmon/Operational

    :param cond_field_template: Template for field name used in added condition. Usually some static
        field name.
    :type cond_field_template: str
    :param cond_value_template: Template for value used in added condition. Usually contains source name.
    :type cond_value_template: str
    :param identifier_template: Template for processing item identifier. Usually, the defaults are
        fine. Should contain service placeholder if changed.
    :type identifier_template: str
    :return: List of ProcessingItem that can be used in the items attribute of a ProcessingPipeline
        object. Usually, an additional field name mapping between the Sigma taxonomy and the target
        system field names is required.
    :rtype: List[ProcessingItem]
    """
    return [
        ProcessingItem(
            identifier=identifier_template.format(service=service, source=source),
            transformation=(
                AddConditionTransformation(
                    {  # source is list
                        cond_field_template.format(service=service, source=source): [
                            cond_value_template.format(service=service, source=source_item)
                            for source_item in source
                        ]
                    }
                )
                if isinstance(source, list)
                else AddConditionTransformation(
                    {  # source is plain string
                        cond_field_template.format(
                            service=service, source=source
                        ): cond_value_template.format(service=service, source=source)
                    }
                )
            ),
            rule_conditions=[logsource_windows(service)],
        )
        for service, source in windows_logsource_mapping.items()
    ]



from collections import UserDict, defaultdict
from dataclasses import dataclass, field
import sigma
from typing import List, Optional, Set, Union


@dataclass
class ProcessingItemTrackingMixin:
    """
    Provides attributes and methods for tracking processing items applied to Sigma rule objects
    like detection items and conditions.
    """

    applied_processing_items: Set[str] = field(init=False, compare=False, default_factory=set)

    def add_applied_processing_item(
        self, processing_item: Optional["sigma.processing.pipeline.ProcessingItem"]
    ):
        """Add identifier of processing item to set of applied processing items."""
        if processing_item is not None and processing_item.identifier is not None:
            self.applied_processing_items.add(processing_item.identifier)

    def was_processed_by(self, processing_item_id: str) -> bool:
        """Determines if detection item was processed by a processing item with the given id."""
        return processing_item_id in self.applied_processing_items


class FieldMappingTracking(UserDict):
    """
    Tracking class for field mappings. Tracks initial field name to finally mapped name after a
    processing pipeline was applied. Each key maps the source field to a set of target fields.

    Currently this class is intentionally only used to track field mappings in detection items and
    the fields list is excluded from it. This might change in the future depending on use cases.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_fields = defaultdict(set)  # Create reverse mapping

    def add_mapping(self, source: str, target: Union[str, List[str]]) -> None:
        """
        This method must be invoked for each field name mapping applied in a processing pipeline to
        get a precise result of the final mapping.
        """
        if not isinstance(target, list):  # Ensure that the target is a list.
            target = [target]

        if source in self.target_fields:  # Source field was already mapping target.
            # Replace each occurrence of a mapping to the source with the target field.
            for source_field in self.target_fields[source]:
                target_set = self[source_field]
                target_set.remove(source)
                target_set.update(target)

            # Update reverse mapping: remove source and add new target
            del self.target_fields[source]
            for t in target:
                self.target_fields[t].add(source_field)

        if source not in self:
            self[source] = set(target)
        else:
            self[source].update(target)
        for t in target:
            self.target_fields[t].add(source)



from dataclasses import dataclass, field
from typing import Callable, Dict, Iterable, List, Optional, Union, IO
from pathlib import Path

from sigma.rule import SigmaRule
from sigma.exceptions import SigmaCollectionError, SigmaError, SigmaRuleLocation
import yaml


@dataclass
class SigmaCollection:
    """Collection of Sigma rules"""

    rules: List[SigmaRule]
    errors: List[SigmaError] = field(default_factory=list)
    ids_to_rules: Dict[str, SigmaRule] = field(init=False, repr=False, hash=False, compare=False)

    def __post_init__(self):
        """
        Map rule identifiers to rules.
        """
        self.ids_to_rules = {str(rule.id): rule for rule in self.rules if rule.id is not None}

    @classmethod
    def from_dicts(
        cls,
        rules: List[dict],
        collect_errors: bool = False,
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaCollection":
        """
        Generate a rule collection from list of dicts containing parsed YAML content.

        If the collect_errors parameters is set, exceptions are not raised while parsing but collected
        in the errors property individually for each Sigma rule and the whole SigmaCollection.
        """
        errors = []
        parsed_rules = list()
        prev_rule = None
        global_rule = dict()

        for i, rule in zip(range(1, len(rules) + 1), rules):
            if isinstance(
                rule, SigmaRule
            ):  # Included rules are already parsed, skip collection action processing
                parsed_rules.append(rule)
                rule.source = source
            else:
                action = rule.get("action")
                if (
                    action is None
                ):  # no action defined: merge with global rule and handle as simple rule
                    parsed_rules.append(
                        SigmaRule.from_dict(
                            deep_dict_update(rule, global_rule), collect_errors, source
                        )
                    )
                    prev_rule = rule
                elif action == "global":  # set global rule template
                    del rule["action"]
                    global_rule = rule
                    prev_rule = global_rule
                elif action == "reset":  # remove global rule
                    global_rule = dict()
                elif (
                    action == "repeat"
                ):  # add content of current rule to previous rule and parse it
                    prev_rule = deep_dict_update(prev_rule, rule)
                    parsed_rules.append(SigmaRule.from_dict(prev_rule, collect_errors, source))
                else:
                    exception = SigmaCollectionError(
                        f"Unknown Sigma collection action '{ action }' in rule { i }",
                        source=source,
                    )
                    if collect_errors:
                        errors.append(exception)
                    else:
                        raise exception

        return cls(parsed_rules, errors)

    @classmethod
    def from_yaml(
        cls,
        yaml_str: Union[bytes, str, IO],
        collect_errors: bool = False,
        source: Optional[SigmaRuleLocation] = None,
    ) -> "SigmaCollection":
        """
        Generate a rule collection from a string containing one or multiple YAML documents.

        If the collect_errors parameters is set, exceptions are not raised while parsing but collected
        in the errors property individually for each Sigma rule and the whole SigmaCollection.
        """
        return cls.from_dicts(list(yaml.safe_load_all(yaml_str)), collect_errors, source)

    @classmethod
    def resolve_paths(
        cls,
        inputs: List[Union[str, Path]],
        recursion_pattern: str = "**/*.yml",
    ) -> Iterable[Path]:
        """
        Resolve list of paths *inputs* that can contain files as well as directories into a flat list of
        files matching *resursion_pattern*.
        """
        paths = (  # Normalize all inputs into paths
            input if isinstance(input, Path) else Path(input) for input in inputs
        )
        paths = (  # Recurse into directories if provided
            path.glob(recursion_pattern) if path.is_dir() else (path,) for path in paths
        )
        return (subpath for subpaths in paths for subpath in subpaths)  # Flatten the list

    @classmethod
    def load_ruleset(
        cls,
        inputs: List[Union[str, Path]],
        collect_errors: bool = False,
        on_beforeload: Optional[Callable[[Path], Optional[Path]]] = None,
        on_load: Optional[Callable[[Path, "SigmaCollection"], Optional["SigmaCollection"]]] = None,
        recursion_pattern: str = "**/*.yml",
    ) -> "SigmaCollection":
        """
        Load a ruleset from a list of files or directories and construct a :class:`SigmaCollection`
        object.

        :param inputs: List of strings and :class:`pathlib.Path` objects that reference files or
        directories that should be loaded.
        :param collect_errors: parse or verification errors are collected in :class:`SigmaRule`
        objects instead of raising them immediately. Defaults to ``False``.
        :param on_beforeload: Optional function that is called for each path to a Sigma rule before the parsing and
        construction of the :class:`SigmaCollection` object is done. The path returned by this function is
        used as input. A rule path is skipped if ``None`` is returned.
        :param on_load: Optional function that is called after the :class:`SigmaCollection` was
        constructed from the path. The path and the SigmaCollection object are passed to this
        function and it is expected to return a :class:`SigmaCollection` object that is merged in
        the collection of the ruleset or ``None`` if the generated collection should be skipped.
        :param recursion_pattern: Pattern used to recurse into directories, defaults to ``**/*.yml``.

        :return: :class:`SigmaCollection` of all sigma rules contained in given paths.
        """
        if not isinstance(inputs, Iterable) or isinstance(inputs, str):
            raise TypeError(
                "Parameter 'inputs' must be list of strings or pathlib.Path objects, not "
                + str(type(inputs))
            )

        paths = cls.resolve_paths(inputs, recursion_pattern)
        sigma_collections = list()
        for path in paths:
            if (
                on_beforeload is not None
            ):  # replace path with return value of on_beforeload function if provided
                path = on_beforeload(path)
            if path is not None:  # Skip if path is None
                sigma_collection = SigmaCollection.from_yaml(
                    path.open(encoding="utf-8"), collect_errors, SigmaRuleLocation(path)
                )
                if (
                    on_load is not None
                ):  # replace SigmaCollection generated from file content with the return value from on_load function if provided
                    sigma_collection = on_load(path, sigma_collection)
                if sigma_collection is not None:  # Skip if nothing
                    sigma_collections.append(sigma_collection)

        # Finally merge all SigmaCollection's and return the result
        return cls.merge(sigma_collections)

    @classmethod
    def merge(cls, collections: Iterable["SigmaCollection"]) -> "SigmaCollection":
        """Merge multiple SigmaCollection objects into one and return it."""
        return cls(
            rules=[rule for collection in collections for rule in collection],
            errors=[error for collection in collections for error in collection.errors],
        )

    def __iter__(self):
        return iter(self.rules)

    def __len__(self):
        return len(self.rules)

    def __getitem__(self, i: Union[int, str]):
        if isinstance(i, int):
            return self.rules[i]
        else:
            return self.ids_to_rules[i]


def deep_dict_update(dest, src):
    for k, v in src.items():
        if isinstance(v, dict):
            dest[k] = deep_dict_update(dest.get(k, {}), v)
        else:
            dest[k] = v
    return dest



from abc import abstractmethod
from dataclasses import dataclass, field
import json
import re
from typing import Any, Dict, List, Optional, Union
import sigma
from sigma.processing.templates import TemplateBase
from sigma.processing.transformations import Transformation
from sigma.rule import SigmaRule


@dataclass
class QueryPostprocessingTransformation(Transformation):
    """Query post processing transformation base class."""

    processing_item: Optional["sigma.processing.pipeline.QueryPostprocessingItem"] = field(
        init=False, compare=False, default=None
    )

    @abstractmethod
    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: Any
    ) -> Any:
        """Applies post-processing transformation to arbitrary typed query.

        :param pipeline: Processing pipeline this transformation was contained.
        :type pipeline: sigma.processing.pipeline.ProcessingPipeline
        :param rule: Sigma rule that is associated with the generated query.
        :type rule: SigmaRule
        :param query: Query that should be transformed.
        :type query: Any
        :return: Transformed query.
        :rtype: Any
        """
        super().apply(pipeline, rule)  # tracking of applied rules and assigning self.pipeline


@dataclass
class EmbedQueryTransformation(QueryPostprocessingTransformation):
    """Embeds a query between a given prefix and suffix. Only applicable to string queries."""

    prefix: Optional[str] = None
    suffix: Optional[str] = None

    def __post_init__(self):
        self.prefix = self.prefix or ""
        self.suffix = self.suffix or ""

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str
    ) -> str:
        super().apply(pipeline, rule, query)
        return self.prefix + query + self.suffix


@dataclass
class QuerySimpleTemplateTransformation(QueryPostprocessingTransformation):
    """Replace query with template that can refer to the following placeholders:
    * query: the postprocessed query.
    * rule: the Sigma rule including all its attributes like rule.title.
    * pipeline: the Sigma processing pipeline where this transformation is applied including all
      current state information in pipeline.state.

    The Python format string syntax (str.format()) is used.
    """

    template: str

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str
    ) -> str:
        return self.template.format(
            query=query,
            rule=rule,
            pipeline=pipeline,
        )


@dataclass
class QueryTemplateTransformation(QueryPostprocessingTransformation, TemplateBase):
    """Apply Jinja2 template provided as template object variable to a query. The following
    variables are available in the context:

    * query: the postprocessed query.
    * rule: the Sigma rule including all its attributes like rule.title.
    * pipeline: the Sigma processing pipeline where this transformation is applied including all
      current state information in pipeline.state.

    if *path* is given, *template* is considered as a relative path to a template file below the
    specified path. If it is not provided, the template is specified as plain string. *autoescape*
    controls the Jinja2 HTML/XML auto-escaping.
    """

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str
    ) -> str:
        return self.j2template.render(query=query, rule=rule, pipeline=pipeline)


@dataclass
class EmbedQueryInJSONTransformation(QueryPostprocessingTransformation):
    """Embeds a query into a JSON structure defined as string. the placeholder value %QUERY% is
    replaced with the query."""

    json_template: str

    def _replace_placeholder(
        self, v: Union[Dict, List, str, int, float], query: str
    ) -> Union[Dict, List, str, int, float]:
        if isinstance(v, dict):
            return {k: self._replace_placeholder(v, query) for k, v in v.items()}
        elif isinstance(v, list):
            return [self._replace_placeholder(i, query) for i in v]
        elif isinstance(v, str) and v == "%QUERY%":
            return query
        else:
            return v

    def __post_init__(self):
        self.parsed_json = json.loads(self.json_template)

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str
    ):
        super().apply(pipeline, rule, query)
        return json.dumps(self._replace_placeholder(self.parsed_json, query))


@dataclass
class ReplaceQueryTransformation(QueryPostprocessingTransformation):
    """Replace query part specified by regular expression with a given string."""

    pattern: str
    replacement: str

    def __post_init__(self):
        self.re = re.compile(self.pattern)

    def apply(
        self, pipeline: "sigma.processing.pipeline.ProcessingPipeline", rule: SigmaRule, query: str
    ):
        super().apply(pipeline, rule, query)
        return self.re.sub(self.replacement, query)


query_postprocessing_transformations = {
    "embed": EmbedQueryTransformation,
    "simple_template": QuerySimpleTemplateTransformation,
    "template": QueryTemplateTransformation,
    "json": EmbedQueryInJSONTransformation,
    "replace": ReplaceQueryTransformation,
}



from dataclasses import dataclass
from typing import Optional

from jinja2 import Environment, FileSystemLoader, Template


@dataclass
class TemplateBase:
    """Base class for Jinja template postprocessors and finalizers."""

    template: str
    path: Optional[str] = None
    autoescape: bool = False

    def __post_init__(self):
        if self.path is None:
            self.j2template = Template(self.template, autoescape=self.autoescape)
        else:
            env = Environment(autoescape=self.autoescape, loader=FileSystemLoader(self.path))
            self.j2template = env.get_template(self.template)



from collections import Counter
from dataclasses import dataclass
from typing import ClassVar, List, Set
from sigma.rule import SigmaRule, SigmaRuleTag
from sigma.validators.base import (
    SigmaRuleValidator,
    SigmaTagValidator,
    SigmaValidationIssue,
    SigmaValidationIssueSeverity,
)
from sigma.data.mitre_attack import (
    mitre_attack_tactics,
    mitre_attack_techniques,
    mitre_attack_intrusion_sets,
    mitre_attack_software,
)
import re


@dataclass
class InvalidATTACKTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid MITRE ATT&CK tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class ATTACKTagValidator(SigmaTagValidator):
    """Check for usage of valid MITRE ATT&CK tags."""

    def __init__(self) -> None:
        self.allowed_tags = (
            {tactic.lower().replace("-", "_") for tactic in mitre_attack_tactics.values()}
            .union({technique.lower() for technique in mitre_attack_techniques.keys()})
            .union({intrusion_set.lower() for intrusion_set in mitre_attack_intrusion_sets})
            .union({software.lower() for software in mitre_attack_software})
        )

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        if tag.namespace == "attack" and tag.name not in self.allowed_tags:
            return [InvalidATTACKTagIssue([self.rule], tag)]
        return []


@dataclass
class InvalidTLPTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid TLP tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class TLPTagValidatorBase(SigmaTagValidator):
    """Base class for TLP tag validation"""

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        if tag.namespace == "tlp" and tag.name not in self.allowed_tags:
            return [InvalidTLPTagIssue([self.rule], tag)]
        return []


class TLPv1TagValidator(TLPTagValidatorBase):
    """Validation of TLP tags according to old version 1 standard."""

    allowed_tags: Set[str] = {
        "white",
        "green",
        "amber",
        "red",
    }


class TLPv2TagValidator(TLPTagValidatorBase):
    """Validation of TLP tags according to version 2 standard."""

    allowed_tags: Set[str] = {
        "clear",
        "green",
        "amber",
        "amber+strict",
        "red",
    }


class TLPTagValidator(TLPTagValidatorBase):
    """Validation of TLP tags from all versions of the TLP standard."""

    allowed_tags: Set[str] = TLPv1TagValidator.allowed_tags.union(TLPv2TagValidator.allowed_tags)


@dataclass
class DuplicateTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "The same tag appears multiple times"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class DuplicateTagValidator(SigmaRuleValidator):
    """Validate rule tag uniqueness."""

    def validate(self, rule: SigmaRule) -> List[SigmaValidationIssue]:
        tags = Counter(rule.tags)
        return [DuplicateTagIssue([rule], tag) for tag, count in tags.items() if count > 1]


@dataclass
class InvalidCVETagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid CVE tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class CVETagValidator(SigmaTagValidator):
    """Validate rule CVE tag"""

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        tags_pattern = re.compile(r"\d+\.\d+$")
        if tag.namespace == "cve" and tags_pattern.match(tag.name) is None:
            return [InvalidCVETagIssue([self.rule], tag)]
        return []


@dataclass
class InvalidDetectionTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid detection tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class DetectionTagValidator(SigmaTagValidator):
    """Validate rule detection tag"""

    allowed_tags = {"dfir", "emerging_threats", "threat_hunting"}

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        if tag.namespace == "detection" and tag.name not in self.allowed_tags:
            return [InvalidDetectionTagIssue([self.rule], tag)]
        return []


@dataclass
class InvalidCARTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid CAR tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class CARTagValidator(SigmaTagValidator):
    """Validate rule CAR tag"""

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        tags_pattern = re.compile(r"\d{4}-\d{2}-\d{3}$")
        if tag.namespace == "car" and tags_pattern.match(tag.name) is None:
            return [InvalidCARTagIssue([self.rule], tag)]
        return []


@dataclass
class InvalidSTPTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid STP tagging"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class STPTagValidator(SigmaTagValidator):
    """Validate rule STP tag"""

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        tags_pattern = re.compile(r"^[1-5]{1}[auk]{0,1}$")
        if tag.namespace == "stp" and tags_pattern.match(tag.name) is None:
            return [InvalidSTPTagIssue([self.rule], tag)]
        return []


@dataclass
class InvalidNamespaceTagIssue(SigmaValidationIssue):
    description: ClassVar[str] = "Invalid tagging name"
    severity: ClassVar[SigmaValidationIssueSeverity] = SigmaValidationIssueSeverity.MEDIUM
    tag: SigmaRuleTag


class NamespaceTagValidator(SigmaTagValidator):
    """Validate rule tag name"""

    allowed_namespace = {"attack", "car", "cve", "detection", "stp"}

    def validate_tag(self, tag: SigmaRuleTag) -> List[SigmaValidationIssue]:
        if tag.namespace not in self.allowed_namespace:
            return [InvalidNamespaceTagIssue([self.rule], tag)]
        return []
